import asyncio
from datetime import datetime
from decimal import Decimal
from typing import Any, Callable, Dict, List, Optional, Tuple

from bidict import bidict

from hummingbot.connector.constants import s_decimal_NaN
from hummingbot.connector.exchange.bitstamp import (
    bitstamp_constants as CONSTANTS,
    bitstamp_utils,
    bitstamp_web_utils as web_utils,
)
from hummingbot.connector.exchange.bitstamp.bitstamp_api_order_book_data_source import BitstampAPIOrderBookDataSource
from hummingbot.connector.exchange.bitstamp.bitstamp_api_user_stream_data_source import BitstampAPIUserStreamDataSource
from hummingbot.connector.exchange.bitstamp.bitstamp_auth import BitstampAuth
from hummingbot.connector.exchange_py_base import ExchangePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderState, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase, TradeFeeSchema
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.connections.data_types import RESTMethod
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class BitstampExchange(ExchangePyBase):
    UPDATE_ORDER_STATUS_MIN_INTERVAL = 10.0

    web_utils = web_utils

    def __init__(self,
                 bitstamp_api_key: str,
                 bitstamp_api_secret: str,
                 balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
                 rate_limits_share_pct: Decimal = Decimal("100"),
                 trading_pairs: Optional[List[str]] = None,
                 trading_required: bool = True,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN,
                 time_provider: Optional[Callable] = None,
                 ):
        self.api_key = bitstamp_api_key
        self.secret_key = bitstamp_api_secret
        self._trading_pairs = trading_pairs
        self._trading_required = trading_required
        self._domain = domain
        self._time_provider = time_provider
        self._last_trades_poll_bitstamp_timestamp = 1.0

        super().__init__(balance_asset_limit, rate_limits_share_pct)
        self._real_time_balance_update = False
        self._trading_fees = {}

    @staticmethod
    def bitstamp_order_type(order_type: OrderType) -> str:
        return order_type.name.upper()

    @staticmethod
    def to_hb_order_type(bitstamp_type: str) -> OrderType:
        return OrderType[bitstamp_type]

    @property
    def authenticator(self):
        return BitstampAuth(
            api_key=self.api_key,
            secret_key=self.secret_key,
            time_provider=self._time_synchronizer)

    @property
    def name(self) -> str:
        return "bitstamp"

    @property
    def rate_limits_rules(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self):
        return self._domain

    @property
    def client_order_id_max_length(self):
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self):
        return CONSTANTS.HBOT_ORDER_ID_PREFIX

    @property
    def trading_rules_request_path(self):
        return CONSTANTS.EXCHANGE_INFO_PATH_URL

    @property
    def trading_pairs_request_path(self):
        return CONSTANTS.EXCHANGE_INFO_PATH_URL

    @property
    def check_network_request_path(self):
        return CONSTANTS.STATUS_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    def supported_order_types(self):
        return [OrderType.LIMIT, OrderType.LIMIT_MAKER, OrderType.MARKET]

    async def get_all_pairs_prices(self) -> List[Dict[str, str]]:
        pairs_prices = await self._api_get(path_url=CONSTANTS.CURRENCIES_URL)
        return pairs_prices

    def convert_from_exchange_trading_pair(self, exchange_trading_pair: str) -> Optional[str]:
        try:
            base_asset, quote_asset = exchange_trading_pair.split("/")
        except Exception as e:
            raise ValueError(f"Error parsing the trading pair {exchange_trading_pair}: {e}")

        return f"{base_asset}-{quote_asset}"

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        return CONSTANTS.TIMESTAMP_ERROR_CODE in str(
            request_exception
        ) and CONSTANTS.TIMESTAMP_ERROR_MESSAGE in str(request_exception)

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        return CONSTANTS.ORDER_NOT_EXIST_ERROR_CODE in str(
            status_update_exception
        ) and CONSTANTS.ORDER_NOT_EXIST_MESSAGE in str(status_update_exception)

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        return CONSTANTS.ORDER_NOT_EXIST_ERROR_CODE in str(
            cancelation_exception
        ) and CONSTANTS.ORDER_NOT_EXIST_MESSAGE in str(cancelation_exception)

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            time_provider=self._time_provider,
            domain=self._domain,
            auth=self._auth)

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return BitstampAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            domain=self.domain,
            api_factory=self._web_assistants_factory)

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return BitstampAPIUserStreamDataSource(
            auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> TradeFeeBase:

        is_maker = is_maker or (order_type is OrderType.LIMIT_MAKER)
        trading_pair = combine_to_hb_trading_pair(base=base_currency, quote=quote_currency)

        trade_fee_schema = self._trading_fees.get(trading_pair)
        if trade_fee_schema:
            fee_percent: Decimal = (
                trade_fee_schema.maker_percent_fee_decimal if is_maker else trade_fee_schema.taker_percent_fee_decimal
            )
            fee = TradeFeeBase.new_spot_fee(
                fee_schema=trade_fee_schema,
                trade_type=order_side,
                percent=fee_percent
            )
        else:
            fee = build_trade_fee(
                self.name,
                is_maker,
                base_currency=base_currency,
                quote_currency=quote_currency,
                order_type=order_type,
                order_side=order_side,
                amount=amount,
                price=price,
            )
        return fee

    async def _place_order(self,
                           order_id: str,
                           trading_pair: str,
                           amount: Decimal,
                           trade_type: TradeType,
                           order_type: OrderType,
                           price: Decimal,
                           **kwargs) -> Tuple[str, float]:
        api_params = {
            "amount": f"{amount:f}",
            "client_order_id": order_id
        }

        side_str = CONSTANTS.SIDE_BUY if trade_type is TradeType.BUY else CONSTANTS.SIDE_SELL
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        if order_type.is_limit_type():
            order_url = f"/{side_str}/{symbol}/"
            api_params["price"] = f"{price:f}"
            if order_type == OrderType.LIMIT_MAKER:
                # Set Maker-Or-Cancel, this ensures that the order is not fully or partially filled when placed.
                # In case it would be, the order is cancelled.
                api_params["moc_order"] = True
        else:
            order_url = f"/{side_str}/market/{symbol}/"

        order_result = await self._api_post(
            path_url=order_url,
            data=api_params,
            is_auth_required=True,
            limit_id=CONSTANTS.ORDER_CREATE_URL_LIMIT_ID
        )

        if order_result.get("status", "") == "error":
            raise IOError(f"Error placing order. Error: {order_result}")

        o_id = str(order_result["id"])
        transact_time = datetime.fromisoformat(order_result["datetime"]).timestamp()

        return o_id, transact_time

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        exchange_order_id = await tracked_order.get_exchange_order_id()

        cancel_response = await self._api_post(
            path_url=f"{CONSTANTS.ORDER_CANCEL_URL}",
            data={"id": exchange_order_id},
            is_auth_required=True
        )
        if cancel_response.get("status", "") == "error":
            raise IOError(f"Error canceling order. Error: {cancel_response}")

        return str(cancel_response.get("id", "")) == exchange_order_id

    async def _format_trading_rules(self, exchange_info: List[Dict[str, Any]]) -> List[TradingRule]:
        retval = []
        for info in filter(bitstamp_utils.is_exchange_information_valid, exchange_info):
            try:
                retval.append(
                    TradingRule(
                        trading_pair=self.convert_from_exchange_trading_pair(info["name"]),
                        min_price_increment=Decimal(f"1e-{info['counter_decimals']}"),
                        min_base_amount_increment=Decimal(f"1e-{info['base_decimals']}"),
                        min_quote_amount_increment=Decimal(f"1e-{info['counter_decimals']}"),
                        min_notional_size=Decimal(info["minimum_order"].split(" ")[0]))
                )
            except Exception:
                self.logger().exception(f"Error parsing the trading pair rule {info}. Skipping.")
        return retval

    async def _update_trading_fees(self):
        """
        Update fees information from the exchange
        """
        trading_fees: List[Dict[str, Any]] = await self._api_post(
            path_url=CONSTANTS.TRADING_FEES_URL,
            is_auth_required=True
        )

        for fee_info in trading_fees:
            try:
                trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=fee_info.get("market"))
            except KeyError:
                continue

            if trading_pair:
                fees = fee_info["fees"]
                self._trading_fees[trading_pair] = TradeFeeSchema(
                    maker_percent_fee_decimal=Decimal(fees["maker"]),
                    taker_percent_fee_decimal=Decimal(fees["taker"])
                )

    async def _user_stream_event_listener(self):
        """
        This functions runs in background continuously processing the events received from the exchange by the user
        stream data source. It keeps reading events from the queue until the task is interrupted.
        The events received are balance updates, order updates and trade events.
        """
        trade_events = {
            CONSTANTS.USER_TRADE,
            CONSTANTS.USER_SELF_TRADE,
        }
        async for event_message in self._iter_user_event_queue():
            try:
                event = event_message.get("event")

                if event in trade_events:
                    self._process_user_stream_trade_event(event, event_message)
                else:
                    self._process_user_stream_order_event(event, event_message)

            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error("Unexpected error in user stream listener loop.", exc_info=True)
                await self._sleep(5.0)

    def _process_user_stream_trade_event(self, event: str, event_message: Dict[str, Any]):
        try:
            event_data = event_message.get("data", {})

            if event == CONSTANTS.USER_TRADE:
                client_order_id = str(event_data.get("client_order_id"))
                order: InFlightOrder = self._order_tracker.all_fillable_orders.get(client_order_id)
                if order is None:
                    self.logger().debug(f"Received event for unknown order ID: {event_message}")
                    return

                fee = TradeFeeBase.new_spot_fee(
                    fee_schema=self.trade_fee_schema(),
                    trade_type=order.trade_type,
                    flat_fees=[TokenAmount(amount=Decimal(event_data["fee"]), token=order.quote_asset)]
                )

                amount = Decimal(event_data["amount"])
                price = Decimal(event_data["price"])
                trade_update = TradeUpdate(
                    trade_id=str(event_data["id"]),
                    client_order_id=order.client_order_id,
                    exchange_order_id=event_data["order_id"],
                    trading_pair=order.trading_pair,
                    fee=fee,
                    fill_base_amount=amount,
                    fill_quote_amount=price * amount,
                    fill_price=price,
                    fill_timestamp=float(event_data["microtimestamp"]) * 1e-3,
                )
                self._order_tracker.process_trade_update(trade_update)
            else:
                # These trades don't incur any fees, but they do offset each other.
                # We register them as regular fills so that the executed amount is updated correctly.

                buy_order_id = str(event_data.get("buy_order_id"))
                sell_order_id = str(event_data.get("sell_order_id"))

                amount = Decimal(event_data["amount"])
                price = Decimal(event_data["price"])

                buy_order: InFlightOrder = self._order_tracker.all_fillable_orders_by_exchange_order_id.get(buy_order_id)
                if buy_order:
                    buy_trade_update = TradeUpdate(
                        trade_id=f"{buy_order_id}-{sell_order_id}",
                        client_order_id=buy_order.client_order_id,
                        exchange_order_id=buy_order_id,
                        trading_pair=buy_order.trading_pair,
                        fee=TradeFeeBase.new_spot_fee(
                            fee_schema=self.trade_fee_schema(),
                            trade_type=buy_order.trade_type,
                            flat_fees=TokenAmount(amount=Decimal(0), token=buy_order.quote_asset)),
                        fill_base_amount=amount,
                        fill_quote_amount=price * amount,
                        fill_price=price,
                        fill_timestamp=float(event_data["timestamp"])
                    )
                    self._order_tracker.process_trade_update(buy_trade_update)

                sell_order: InFlightOrder = self._order_tracker.all_fillable_orders_by_exchange_order_id.get(sell_order_id)
                if sell_order:
                    sell_trade_update = TradeUpdate(
                        trade_id=f"{buy_order_id}-{sell_order_id}",
                        client_order_id=sell_order.client_order_id,
                        exchange_order_id=sell_order_id,
                        trading_pair=sell_order.trading_pair,
                        fee=TradeFeeBase.new_spot_fee(
                            fee_schema=self.trade_fee_schema(),
                            trade_type=sell_order.trade_type,
                            flat_fees=TokenAmount(amount=Decimal(0), token=sell_order.quote_asset)),
                        fill_base_amount=amount,
                        fill_quote_amount=price * amount,
                        fill_price=price,
                        fill_timestamp=float(event_data["timestamp"]),
                    )
                    self._order_tracker.process_trade_update(sell_trade_update)

        except Exception as e:
            raise ValueError(f"Error parsing the user stream trade event {event_message}: {e}")

    def _process_user_stream_order_event(self, event: str, event_message: Dict[str, Any]):
        try:
            event_data = event_message.get("data", {})
            client_order_id = str(event_data.get("client_order_id"))
            order: InFlightOrder = self._order_tracker.all_fillable_orders.get(client_order_id)
            if order is None:
                self.logger().debug(f"Received event for unknown order ID: {event_message}")
                return

            # Determine the new state of the order
            new_state = OrderState.OPEN
            if event == CONSTANTS.USER_ORDER_CHANGED and Decimal(event_data["amount_traded"]) > 0:
                new_state = OrderState.PARTIALLY_FILLED
            elif event == CONSTANTS.USER_ORDER_DELETED:
                new_state = OrderState.FILLED if event_data["amount"] == 0 else OrderState.CANCELED

            order_update = OrderUpdate(
                client_order_id=order.client_order_id,
                exchange_order_id=order.exchange_order_id,
                trading_pair=order.trading_pair,
                update_timestamp=float(event_data["datetime"]),
                new_state=new_state,
            )
            self._order_tracker.process_order_update(order_update)
        except Exception as e:
            raise ValueError(f"Error parsing the user stream order event {event_message}: {e}")

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        all_fills_response = await self._api_post(
            path_url=CONSTANTS.ORDER_STATUS_URL,
            data={
                "client_order_id": order.client_order_id,
                "omit_transactions": "false",
            },
            is_auth_required=True
        )

        exchange_order_id = await order.get_exchange_order_id()
        trade_updates = []
        for trade in all_fills_response.get("transactions", []):
            fee = TradeFeeBase.new_spot_fee(
                fee_schema=self.trade_fee_schema(),
                trade_type=order.trade_type,
                flat_fees=[TokenAmount(amount=Decimal(trade["fee"]), token=order.quote_asset)]
            )
            trade_update = TradeUpdate(
                trade_id=str(trade["tid"]),
                client_order_id=order.client_order_id,
                exchange_order_id=exchange_order_id,
                trading_pair=order.trading_pair,
                fee=fee,
                fill_base_amount=Decimal(trade[order.base_asset.lower()]),
                fill_quote_amount=Decimal(trade[order.quote_asset.lower()]),
                fill_price=Decimal(trade["price"]),
                fill_timestamp=datetime.fromisoformat(trade["datetime"]).timestamp(),
            )
            trade_updates.append(trade_update)

        return trade_updates

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        updated_order_data = await self._api_post(
            path_url=CONSTANTS.ORDER_STATUS_URL,
            data={
                "client_order_id": tracked_order.client_order_id,
                "omit_transactions": "true"
            },
            is_auth_required = True
        )

        if updated_order_data.get("status", "") == "error":
            raise IOError(f"Error requesting order status. Error: {updated_order_data}")

        new_state = CONSTANTS.ORDER_STATE[updated_order_data["status"]]
        amount_remaining = Decimal(updated_order_data["amount_remaining"])
        if new_state == OrderState.OPEN and amount_remaining < tracked_order.amount:
            new_state = OrderState.PARTIALLY_FILLED

        order_update = OrderUpdate(
            client_order_id=tracked_order.client_order_id,
            exchange_order_id=updated_order_data["id"],
            trading_pair=tracked_order.trading_pair,
            update_timestamp=datetime.fromisoformat(updated_order_data["datetime"]).timestamp(),
            new_state=new_state,
        )

        return order_update

    async def _update_balances(self):
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()

        balances = await self._api_post(
            path_url=CONSTANTS.ACCOUNT_BALANCES_URL,
            is_auth_required=True
        )

        for balance_entry in balances:
            asset_name = balance_entry["currency"].upper()
            self._account_available_balances[asset_name] = Decimal(balance_entry["available"])
            self._account_balances[asset_name] = Decimal(balance_entry["total"])
            remote_asset_names.add(asset_name)

        asset_names_to_remove = local_asset_names.difference(remote_asset_names)
        for asset_name in asset_names_to_remove:
            del self._account_available_balances[asset_name]
            del self._account_balances[asset_name]

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: List[Dict[str, Any]]):
        mapping = bidict()
        for info in filter(bitstamp_utils.is_exchange_information_valid, exchange_info):
            try:
                mapping[info["url_symbol"]] = self.convert_from_exchange_trading_pair(info["name"])
            except Exception:
                self.logger().error(f"Error parsing trading pair symbol data {info}. Skipping.")

        self._set_trading_pair_symbol_map(mapping)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)

        resp_json = await self._api_get(
            method=RESTMethod.GET,
            path_url=CONSTANTS.TICKER_URL.format(symbol),
            limit_id=CONSTANTS.TICKER_URL_LIMIT_ID
        )

        return float(resp_json["last"])



import asyncio
import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.cube import cube_constants as CONSTANTS, cube_web_utils as web_utils
from hummingbot.connector.exchange.cube.cube_order_book import CubeOrderBook
from hummingbot.connector.exchange.cube.cube_ws_protobufs import market_data_pb2
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.order_book_message import OrderBookMessage
from hummingbot.core.data_type.order_book_row import OrderBookRow
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.utils.async_utils import safe_gather
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSBinaryRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.cube.cube_exchange import CubeExchange


class CubeAPIOrderBookDataSource(OrderBookTrackerDataSource):
    HEARTBEAT_TIME_INTERVAL = 30.0
    TRADE_STREAM_ID = 1
    DIFF_STREAM_ID = 2
    ONE_HOUR = 60 * 60

    _logger: Optional[HummingbotLogger] = None

    def __init__(self,
                 trading_pairs: List[str],
                 connector: 'CubeExchange',
                 api_factory: WebAssistantsFactory,
                 domain: str = CONSTANTS.DEFAULT_DOMAIN):
        super().__init__(trading_pairs)
        self._connector = connector
        self._trade_messages_queue_key = CONSTANTS.TRADE_EVENT_TYPE
        self._diff_messages_queue_key = CONSTANTS.DIFF_EVENT_TYPE
        self._snapshot_messages_queue_key = CONSTANTS.SNAPSHOT_EVENT_TYPE
        self._domain = domain
        self._api_factory = api_factory

    async def get_last_traded_prices(self,
                                     trading_pairs: List[str],
                                     domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        """
        Retrieves a copy of the full order book from the exchange, for a particular trading pair.

        :param trading_pair: the trading pair for which the order book will be retrieved

        :return: the response from the exchange (JSON dictionary)
        """
        params = {
            "mbp": "true",
            "levels": 1000
        }

        try:
            market_id = await self._connector.exchange_market_id_associated_to_pair(trading_pair=trading_pair)
            rest_assistant = await self._api_factory.get_rest_assistant()
            data = await rest_assistant.execute_request(
                url=web_utils.public_rest_url(
                    path_url=CONSTANTS.MARKET_DATA_REQUEST_URL + f"/book/{market_id}/snapshot",
                    domain=self._domain),
                params=params,
                method=RESTMethod.GET,
                throttler_limit_id=CONSTANTS.SNAPSHOT_LM_ID,
            )
        except Exception as e:
            self.logger().error(f"Error fetching order book snapshot for {trading_pair}: {e}")
            return {}

        return data

    async def _subscribe_channels(self, ws: WSAssistant):
        pass

    async def _connected_websocket_assistant(self) -> WSAssistant:
        pass

    async def _connected_websocket_assistant_for_pair(self, trading_pair: str) -> WSAssistant:
        ws: WSAssistant = await self._api_factory.get_ws_assistant()

        market_id = await self._connector.exchange_market_id_associated_to_pair(trading_pair=trading_pair)

        await ws.connect(
            ws_url=f"{CONSTANTS.WSS_MARKET_DATA_URL.get(self._domain)}/book/{market_id}?mbp=true&trades=true",
            ping_timeout=CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)

        self.logger().info(f"Subscribed to public order book for {trading_pair} and trade channels...")

        return ws

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_timestamp: float = snapshot["result"]["lastTransactTime"]

        price_scaler = await self._connector.get_price_scaler(trading_pair)
        quantity_scaler = await self._connector.get_quantity_scaler(trading_pair)

        snapshot_msg: OrderBookMessage = CubeOrderBook.snapshot_message_from_exchange(
            msg=snapshot,
            timestamp=snapshot_timestamp,
            metadata={"trading_pair": trading_pair},
            price_scaler=price_scaler,
            quantity_scaler=quantity_scaler,
        )

        return snapshot_msg

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        trading_pair = raw_message["trading_pair"]
        trades: market_data_pb2.Trades = raw_message["trades"]
        trade: market_data_pb2.Trades.Trade

        price_scaler = await self._connector.get_price_scaler(trading_pair)
        quantity_scaler = await self._connector.get_quantity_scaler(trading_pair)

        for trade in trades.trades:
            msg = {
                "trading_pair": trading_pair,
                "price": price_scaler * trade.price,
                "fill_quantity": quantity_scaler * trade.fill_quantity,
                "transact_time": trade.transact_time,
                "trade_id": trade.tradeId,
                "trade_type": float(
                    TradeType.SELL.value) if trade.aggressing_side == market_data_pb2.Side.ASK else float(
                    TradeType.BUY.value),
                "timestamp": time.time(),
            }

            trade_message = CubeOrderBook.trade_message_from_exchange(msg)
            message_queue.put_nowait(trade_message)

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        trading_pair = raw_message["trading_pair"]
        diff_msg: market_data_pb2.MarketByPriceDiff = raw_message["mbp_diff"]
        # mbp_diff = market_data_pb2.MarketByPriceDiff().From
        # ParseDict(diff_msg, mbp_diff)
        diff: market_data_pb2.MarketByPriceDiff.Diff

        price_scaler = await self._connector.get_price_scaler(trading_pair)
        quantity_scaler = await self._connector.get_quantity_scaler(trading_pair)

        # Catch if diffs is not iterable
        if not hasattr(diff_msg, 'diffs'):
            self.logger().warning(f"Diff message does not contain diffs: {diff_msg}")
            return

        for diff in diff_msg.diffs:
            asks: List[OrderBookRow] = [OrderBookRow(0, 0, 0) for _ in range(0)]
            bids: List[OrderBookRow] = [OrderBookRow(0, 0, 0) for _ in range(0)]
            price = diff.price * price_scaler
            qty = diff.quantity * quantity_scaler
            update_id = int(time.time_ns())

            match diff.op:
                case market_data_pb2.MarketByPriceDiff.REMOVE:
                    if diff.side == market_data_pb2.ASK:
                        row = OrderBookRow(price, 0, update_id)
                        asks.append(row)
                    else:
                        row = OrderBookRow(price, 0, update_id)
                        bids.append(row)
                case market_data_pb2.MarketByPriceDiff.REPLACE:
                    if diff.side == market_data_pb2.ASK:
                        row = OrderBookRow(price, qty, update_id)
                        asks.append(row)
                    else:
                        row = OrderBookRow(price, qty, update_id)
                        bids.append(row)
            msg = {"trading_pair": trading_pair, "update_id": update_id, "bids": bids, "asks": asks}

            order_book_message: OrderBookMessage = CubeOrderBook.diff_message_from_exchange(
                msg, time.time())
            message_queue.put_nowait(order_book_message)

    async def _process_websocket_messages_for_pair(self, websocket_assistant: WSAssistant, trading_pair: str):
        async def handle_heartbeat():
            send_hb = True
            while send_hb:
                await asyncio.sleep(CONSTANTS.WS_HEARTBEAT_TIME_INTERVAL)
                hb = market_data_pb2.Heartbeat(
                    request_id=0,
                    timestamp=time.time_ns(),
                )
                hb_request: WSBinaryRequest = WSBinaryRequest(
                    payload=market_data_pb2.ClientMessage(heartbeat=hb).SerializeToString())
                try:
                    await websocket_assistant.send(hb_request)
                except asyncio.CancelledError:
                    send_hb = False
                except ConnectionError:
                    send_hb = False
                except RuntimeError:
                    send_hb = False

        async def handle_messages():
            data: market_data_pb2.MdMessages
            async for ws_response in websocket_assistant.iter_messages():
                data = market_data_pb2.MdMessages().FromString(ws_response.data)
                if data is not None:  # data will be None when the websocket is disconnected
                    for md_msg in data.messages:
                        field = md_msg.WhichOneof('inner')
                        if field == CONSTANTS.DIFF_EVENT_TYPE:
                            diff_data = md_msg.mbp_diff
                            self._message_queue[CONSTANTS.DIFF_EVENT_TYPE].put_nowait(
                                {"trading_pair": trading_pair, "mbp_diff": diff_data})
                        elif field == CONSTANTS.TRADE_EVENT_TYPE:
                            trade_data = md_msg.trades
                            self._message_queue[CONSTANTS.TRADE_EVENT_TYPE].put_nowait(
                                {"trading_pair": trading_pair, "trades": trade_data})

        tasks = [handle_heartbeat(), handle_messages()]
        await safe_gather(*tasks)

    async def listen_for_subscriptions(self):
        """
        Connects to the trade events and order diffs websocket endpoints and listens to the messages sent by the
        exchange. Each message is stored in its own queue.
        """

        async def handle_subscription(trading_pair):
            ws: Optional[WSAssistant] = None
            while True:
                try:
                    ws: WSAssistant = await self._connected_websocket_assistant_for_pair(trading_pair=trading_pair)
                    await self._process_websocket_messages_for_pair(websocket_assistant=ws, trading_pair=trading_pair)
                except asyncio.CancelledError:
                    raise
                except ConnectionError as connection_exception:
                    self.logger().warning(
                        f"The websocket connection to {trading_pair} was closed ({connection_exception})")
                except Exception:
                    self.logger().exception(
                        "Unexpected error occurred when listening to order book streams. Retrying in 5 seconds...",
                    )
                    await self._sleep(1.0)
                finally:
                    await self._on_order_stream_interruption(websocket_assistant=ws)

        tasks = [handle_subscription(trading_pair) for trading_pair in self._trading_pairs]
        await safe_gather(*tasks)

    async def subscribe_to_trading_pair(self, trading_pair: str) -> bool:
        """Dynamic subscription not supported for this connector."""
        self.logger().warning(
            f"Dynamic subscription not supported for {self.__class__.__name__}"
        )
        return False

    async def unsubscribe_from_trading_pair(self, trading_pair: str) -> bool:
        """Dynamic unsubscription not supported for this connector."""
        self.logger().warning(
            f"Dynamic unsubscription not supported for {self.__class__.__name__}"
        )
        return False



from typing import (
    Dict,
    List,
    Tuple
)

from hummingbot.core.data_type.limit_order cimport LimitOrder
from hummingbot.core.data_type.limit_order import LimitOrder
from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
from hummingbot.strategy.order_tracker cimport OrderTracker

NaN = float("nan")


cdef class PureMarketMakingOrderTracker(OrderTracker):
    # ETH confirmation requirement of Binance has shortened to 12 blocks as of 7/15/2019.
    # 12 * 15 / 60 = 3 minutes
    SHADOW_MAKER_ORDER_KEEP_ALIVE_DURATION = 60.0 * 3

    def __init__(self):
        super().__init__()

    @property
    def active_limit_orders(self) -> List[Tuple[ConnectorBase, LimitOrder]]:
        limit_orders = []
        for market_pair, orders_map in self._tracked_limit_orders.items():
            for limit_order in orders_map.values():
                limit_orders.append((market_pair.market, limit_order))
        return limit_orders

    @property
    def shadow_limit_orders(self) -> List[Tuple[ConnectorBase, LimitOrder]]:
        limit_orders = []
        for market_pair, orders_map in self._shadow_tracked_limit_orders.items():
            for limit_order in orders_map.values():
                limit_orders.append((market_pair.market, limit_order))
        return limit_orders

    @property
    def market_pair_to_active_orders(self) -> Dict[MarketTradingPairTuple, List[LimitOrder]]:
        market_pair_to_orders = {}
        market_pairs = self._tracked_limit_orders.keys()
        for market_pair in market_pairs:
            maker_orders = []
            for limit_order in self._tracked_limit_orders[market_pair].values():
                maker_orders.append(limit_order)
            market_pair_to_orders[market_pair] = maker_orders
        return market_pair_to_orders



from decimal import Decimal
from typing import Dict

from hummingbot.connector.utils import combine_to_hb_trading_pair, split_hb_trading_pair
from hummingbot.core.gateway.utils import unwrap_token_symbol


def find_rate(prices: Dict[str, Decimal], pair: str) -> Decimal:
    '''
    Finds exchange rate for a given trading pair from a dictionary of prices
    For example, given prices of {"HBOT-USDT": Decimal("100"), "AAVE-USDT": Decimal("50"), "USDT-GBP": Decimal("0.75")}
    A rate for USDT-HBOT will be 1 / 100
    A rate for HBOT-AAVE will be 100 / 50
    A rate for AAVE-HBOT will be 50 / 100
    A rate for HBOT-GBP will be 100 * 0.75
    :param prices: The dictionary of trading pairs and their prices
    :param pair: The trading pair
    '''
    if pair in prices:
        return prices[pair]
    base, quote = split_hb_trading_pair(trading_pair=pair)
    base = unwrap_token_symbol(base)
    quote = unwrap_token_symbol(quote)
    if base == quote:
        return Decimal("1")
    reverse_pair = combine_to_hb_trading_pair(base=quote, quote=base)
    if reverse_pair in prices:
        return Decimal("1") / prices[reverse_pair]
    base_prices = {k: v for k, v in prices.items() if k.startswith(f"{base}-")}
    for base_pair, proxy_price in base_prices.items():
        link_quote = split_hb_trading_pair(base_pair)[1]
        link_pair = combine_to_hb_trading_pair(base=link_quote, quote=quote)
        if link_pair in prices:
            return proxy_price * prices[link_pair]
        common_denom_pair = combine_to_hb_trading_pair(base=quote, quote=link_quote)
        if common_denom_pair in prices:
            return proxy_price / prices[common_denom_pair]



from prompt_toolkit.layout.containers import Window
from prompt_toolkit.buffer import Buffer
from typing import Optional


def scroll_down(event, window: Optional[Window] = None, buffer: Optional[Buffer] = None):
    w = window or event.app.layout.current_window
    b = buffer or event.app.current_buffer

    if w and w.render_info:
        info = w.render_info
        ui_content = info.ui_content

        # Height to scroll.
        scroll_height = info.window_height // 2

        # Calculate how many lines is equivalent to that vertical space.
        y = b.document.cursor_position_row + 1
        height = 0
        while y < ui_content.line_count:
            line_height = info.get_height_for_line(y)

            if height + line_height < scroll_height:
                height += line_height
                y += 1
            else:
                break

        b.cursor_position = b.document.translate_row_col_to_index(y, 0)


def scroll_up(event, window: Optional[Window] = None, buffer: Optional[Buffer] = None):
    w = window or event.app.layout.current_window
    b = buffer or event.app.current_buffer

    if w and w.render_info:
        info = w.render_info

        # Height to scroll.
        scroll_height = info.window_height // 2

        # Calculate how many lines is equivalent to that vertical space.
        y = max(0, b.document.cursor_position_row - 1)
        height = 0
        while y > 0:
            line_height = info.get_height_for_line(y)

            if height + line_height < scroll_height:
                height += line_height
                y -= 1
            else:
                break

        b.cursor_position = b.document.translate_row_col_to_index(y, 0)



#!/usr/bin/env python

from prompt_toolkit.application.current import get_app
from prompt_toolkit.filters import is_searching, to_filter
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.search import SearchDirection, do_incremental_search, start_search, stop_search

from hummingbot.client.ui.scroll_handlers import scroll_down, scroll_up
from hummingbot.client.ui.style import reset_style
from hummingbot.core.utils.async_utils import safe_ensure_future


def load_key_bindings(hb) -> KeyBindings:
    bindings = KeyBindings()

    @bindings.add("c-c", "c-c")
    def exit_(event):
        hb.app.log("\n[Double CTRL + C] keyboard exit")
        safe_ensure_future(hb.exit_loop())

    @bindings.add("c-x")
    def stop_configuration(event):
        hb.app.log("\n[CTRL + X] Exiting config...")
        hb.app.to_stop_config = True
        hb.app.pending_input = " "
        hb.app.input_event.set()
        hb.app.change_prompt(prompt=">>> ")
        hb.placeholder_mode = False
        hb.app.hide_input = False

    @bindings.add("c-s")
    def status(event):
        hb.app.log("\n[CTRL + S] Status")
        hb.status()

    @bindings.add("c-f", filter=to_filter(not is_searching()))
    def do_find(event):
        start_search(hb.app.log_field.control)

    @bindings.add("c-f", filter=is_searching)
    def do_exit_find(event):
        stop_search()
        get_app().layout.focus(hb.app.input_field.control)
        get_app().invalidate()

    @bindings.add("c-z")
    def do_undo(event):
        get_app().layout.current_buffer.undo()

    @bindings.add("c-m", filter=is_searching)
    def do_find_next(event):
        do_incremental_search(direction=SearchDirection.FORWARD)

    @bindings.add("c-c")
    def do_copy(event):
        data = get_app().layout.current_buffer.copy_selection()
        get_app().clipboard.set_data(data)

    @bindings.add("c-v")
    def do_paste(event):
        get_app().layout.current_buffer.paste_clipboard_data(get_app().clipboard.get_data())

    @bindings.add("c-a")
    def do_select_all(event):
        current_buffer = get_app().layout.current_buffer
        current_buffer.cursor_position = 0
        current_buffer.start_selection()
        current_buffer.cursor_position = len(current_buffer.text)

    @bindings.add("c-d")
    def scroll_down_output(event):
        event.app.layout.current_window = hb.app.output_field.window
        event.app.layout.focus = hb.app.output_field.buffer
        scroll_down(event, hb.app.output_field.window, hb.app.output_field.buffer)
        event.app.layout.current_window = hb.app.input_field.window
        event.app.layout.focus = hb.app.input_field.buffer

    @bindings.add("c-e")
    def scroll_up_output(event):
        event.app.layout.current_window = hb.app.output_field.window
        event.app.layout.focus = hb.app.output_field.buffer
        scroll_up(event, hb.app.output_field.window, hb.app.output_field.buffer)
        event.app.layout.current_window = hb.app.input_field.window
        event.app.layout.focus = hb.app.input_field.buffer

    @bindings.add("escape")
    def stop_live_update(event):
        hb.app.live_updates = False

    @bindings.add("c-r")
    def do_reset_style(event):
        hb.app.app.style = reset_style(hb.client_config_map)

    @bindings.add("c-t")
    def toggle_logs(event):
        hb.app.toggle_right_pane()

    @bindings.add('c-b')
    def do_tab_navigate_left(event):
        hb.app.tab_navigate_left()

    @bindings.add('c-n')
    def do_tab_navigate_right(event):
        hb.app.tab_navigate_right()

    return bindings



from typing import List, Tuple

import hummingbot.client.settings as settings
from hummingbot.strategy.cross_exchange_market_making.cross_exchange_market_making import (
    CrossExchangeMarketMakingStrategy,
    LogOption,
)
from hummingbot.strategy.maker_taker_market_pair import MakerTakerMarketPair
from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple


async def start(self):
    c_map = self.strategy_config_map
    maker_market = c_map.maker_market.lower()
    taker_market = c_map.taker_market.lower()
    raw_maker_trading_pair = c_map.maker_market_trading_pair
    raw_taker_trading_pair = c_map.taker_market_trading_pair
    status_report_interval = self.client_config_map.strategy_report_interval

    # Post validation logic moved from pydantic config
    settings.required_exchanges.add(c_map.maker_market)
    settings.required_exchanges.add(c_map.taker_market)

    first_base, first_quote = c_map.maker_market_trading_pair.split("-")
    second_base, second_quote = c_map.taker_market_trading_pair.split("-")
    if first_base != second_base or first_quote != second_quote:
        settings.required_rate_oracle = True
        settings.rate_oracle_pairs = []
        if first_base != second_base:
            settings.rate_oracle_pairs.append(f"{second_base}-{first_base}")
        if first_quote != second_quote:
            settings.rate_oracle_pairs.append(f"{second_quote}-{first_quote}")
    else:
        settings.required_rate_oracle = False
        settings.rate_oracle_pairs = []

    try:
        maker_trading_pair: str = raw_maker_trading_pair
        taker_trading_pair: str = raw_taker_trading_pair
        maker_base, maker_quote = maker_trading_pair.split("-")
        taker_base, taker_quote = taker_trading_pair.split("-")
        maker_assets: Tuple[str, str] = (maker_base, maker_quote)
        taker_assets: Tuple[str, str] = (taker_base, taker_quote)
    except ValueError as e:
        self.notify(str(e))
        return

    market_names: List[Tuple[str, List[str]]] = [
        (maker_market, [maker_trading_pair]),
        (taker_market, [taker_trading_pair]),
    ]

    await self.initialize_markets(market_names)
    maker_data = [self.markets[maker_market], maker_trading_pair] + list(maker_assets)
    taker_data = [self.markets[taker_market], taker_trading_pair] + list(taker_assets)
    maker_market_trading_pair_tuple = MarketTradingPairTuple(*maker_data)
    taker_market_trading_pair_tuple = MarketTradingPairTuple(*taker_data)
    self.market_trading_pair_tuples = [maker_market_trading_pair_tuple, taker_market_trading_pair_tuple]
    self.market_pair = MakerTakerMarketPair(maker=maker_market_trading_pair_tuple,
                                            taker=taker_market_trading_pair_tuple)

    strategy_logging_options = (
        LogOption.CREATE_ORDER,
        LogOption.ADJUST_ORDER,
        LogOption.MAKER_ORDER_FILLED,
        LogOption.REMOVING_ORDER,
        LogOption.STATUS_REPORT,
        LogOption.MAKER_ORDER_HEDGED
    )
    self.strategy = CrossExchangeMarketMakingStrategy()
    self.strategy.init_params(
        config_map=c_map,
        market_pairs=[self.market_pair],
        status_report_interval=status_report_interval,
        logging_options=strategy_logging_options,
        hb_app_notification=True,
    )



from hummingbot.core.data_type.common import OrderType, PositionMode
from hummingbot.core.data_type.in_flight_order import OrderState

# Init connector
EXCHANGE_NAME = "okx_perpetual"
REST_API_VERSION = "v5"
HBOT_BROKER_ID = "Hummingbot"
CLIENT_ID_PREFIX = "93027a12dac34fBC"
MAX_ID_LEN = 32

# -------------------------------------------
# BASE URLS
# -------------------------------------------
DEFAULT_DOMAIN = EXCHANGE_NAME
AWS_DOMAIN = "okx_perpetual_aws"
DEMO_DOMAIN = "okx_perpetual_demo"

REST_URLS = {DEFAULT_DOMAIN: "https://www.okx.com",
             AWS_DOMAIN: "https://aws.okx.com",
             DEMO_DOMAIN: "https://www.okx.com"}

ACCOUNT_MODE = "Single-currency margin mode"
# -------------------------------------------
# DATA TYPES
# -------------------------------------------
POSITION_MODE_API_ONEWAY = "net_mode"
POSITION_MODE_API_HEDGE = "long_short_mode"
POSITION_MODE_MAP = {
    PositionMode.ONEWAY: POSITION_MODE_API_ONEWAY,
    PositionMode.HEDGE: POSITION_MODE_API_HEDGE,
}

ORDER_TYPE_MAP = {
    OrderType.LIMIT: "limit",
    OrderType.MARKET: "market",
    OrderType.LIMIT_MAKER: "post_only",
}

GET = "GET"
POST = "POST"
METHOD = "METHOD"
ENDPOINT = "ENDPOINT"

# Order Status
ORDER_STATE = {
    "live": OrderState.OPEN,
    "filled": OrderState.FILLED,
    "partially_filled": OrderState.PARTIALLY_FILLED,
    "canceled": OrderState.CANCELED,
    "mmp_canceled": OrderState.CANCELED,
}

FUNDING_PAYMENT_EXPENSE_SUBTYPE = "173"
FUNDING_PAYMENT_INCOME_SUBTYPE = "174"
FUNDING_PAYMENT_TYPE = "8"
# -------------------------------------------
# WEB SOCKET ENDPOINTS
# -------------------------------------------
WSS_PUBLIC_URLS = {DEFAULT_DOMAIN: f"wss://ws.okx.com:8443/ws/{REST_API_VERSION}/public",
                   AWS_DOMAIN: f"wss://wsaws.okx.com:8443/ws/{REST_API_VERSION}/public",
                   DEMO_DOMAIN: f"wss://wspap.okx.com:8443/ws/{REST_API_VERSION}/public?brokerId=9999"}

WSS_PRIVATE_URLS = {DEFAULT_DOMAIN: f"wss://ws.okx.com:8443/ws/{REST_API_VERSION}/private",
                    AWS_DOMAIN: f"wss://wsaws.okx.com:8443/ws/{REST_API_VERSION}/private",
                    DEMO_DOMAIN: f"wss://wspap.okx.com:8443/ws/{REST_API_VERSION}/private?brokerId=9999"}

WSS_BUSINESS_URLS = {DEFAULT_DOMAIN: f"wss://ws.okx.com:8443/ws/{REST_API_VERSION}/business",
                     AWS_DOMAIN: f"wss://wsaws.okx.com:8443/ws/{REST_API_VERSION}/business",
                     DEMO_DOMAIN: f"wss://wspap.okx.com:8443/ws/{REST_API_VERSION}/business?brokerId=9999"}
SECONDS_TO_WAIT_TO_RECEIVE_MESSAGE = 25
WS_PING_REQUEST = "ping"
WS_PONG_RESPONSE = "pong"

# API ORDER BOOK DATA SOURCE CHANNELS
WS_TRADES_CHANNEL = "trades"
WS_TRADES_ALL_CHANNEL = "trades-all"
WS_MARK_PRICE_CHANNEL = "mark-price"
WS_INDEX_TICKERS_CHANNEL = "index-tickers"
WS_FUNDING_INFO_CHANNEL = "funding-rate"
WS_ORDER_BOOK_400_DEPTH_100_MS_EVENTS_CHANNEL = "books"
WS_ORDER_BOOK_5_DEPTH_100_MS_EVENTS_CHANNEL = "books5"
WS_ORDER_BOOK_1_DEPTH_10_MS_EVENTS_CHANNEL = "bbo-tbt"
WS_INSTRUMENTS_INFO_CHANNEL = "instruments"

# USER STREAM DATA SOURCE CHANNELS
WS_ACCOUNT_CHANNEL = "account"
WS_BALANCE_AND_POSITIONS_CHANNEL = "balance_and_position"
WS_POSITIONS_CHANNEL = "positions"
WS_ORDERS_CHANNEL = "orders"

# -------------------------------------------
# WEB UTILS ENDPOINTS
# The structure is REST_url = {method: GET/POST, endpoint: /api/v5/...} since for the same endpoint you can have
# different methods. This is also useful for rate limit ids.
# -------------------------------------------
# REST API Public Endpoints
REST_LATEST_SYMBOL_INFORMATION = {METHOD: GET,
                                  ENDPOINT: f"/api/{REST_API_VERSION}/market/tickers"}
REST_ORDER_BOOK = {METHOD: GET,
                   ENDPOINT: f"/api/{REST_API_VERSION}/market/books"}
REST_SERVER_TIME = {METHOD: GET,
                    ENDPOINT: f"/api/{REST_API_VERSION}/public/time"}
REST_MARK_PRICE = {METHOD: GET,
                   ENDPOINT: f"/api/{REST_API_VERSION}/public/mark-price"}
REST_INDEX_TICKERS = {METHOD: GET,
                      ENDPOINT: f"/api/{REST_API_VERSION}/market/index-tickers"}
REST_GET_INSTRUMENTS = {METHOD: GET,
                        ENDPOINT: f"/api/{REST_API_VERSION}/public/instruments"}

# REST API Private General Endpoints
REST_GET_WALLET_BALANCE = {METHOD: GET,
                           ENDPOINT: f"/api/{REST_API_VERSION}/account/balance"}
REST_SET_POSITION_MODE = {METHOD: POST,
                          ENDPOINT: f"/api/{REST_API_VERSION}/account/set-position-mode"}

# REST API Private Pair Specific Endpoints
REST_SET_LEVERAGE = {METHOD: POST,
                     ENDPOINT: f"/api/{REST_API_VERSION}/account/set-leverage"}
REST_FUNDING_RATE_INFO = {METHOD: GET,
                          ENDPOINT: f"/api/{REST_API_VERSION}/public/funding-rate"}
REST_GET_POSITIONS = {METHOD: GET,
                      ENDPOINT: f"/api/{REST_API_VERSION}/account/positions"}
REST_PLACE_ACTIVE_ORDER = {METHOD: POST,
                           ENDPOINT: f"/api/{REST_API_VERSION}/trade/order"}
REST_CANCEL_ACTIVE_ORDER = {METHOD: POST,
                            ENDPOINT: f"/api/{REST_API_VERSION}/trade/cancel-order"}
REST_QUERY_ACTIVE_ORDER = {METHOD: GET,
                           ENDPOINT: REST_PLACE_ACTIVE_ORDER[ENDPOINT]}
REST_USER_TRADE_RECORDS = {METHOD: GET,
                           ENDPOINT: f"/api/{REST_API_VERSION}/trade/fills"}
REST_BILLS_DETAILS = {METHOD: GET,
                      ENDPOINT: f"/api/{REST_API_VERSION}/account/bills"}
REST_WS_LOGIN_PATH = {METHOD: GET,
                      ENDPOINT: "/users/self/verify"}


# -------------------------------------------
# RET CODES
# -------------------------------------------

RET_CODE_OK = "0"

RET_CODE_TIMESTAMP_HEADER_MISSING = "50107"
RET_CODE_TIMESTAMP_HEADER_INVALID = "50112"
RET_CODE_PARAMS_ERROR = "51000"
RET_CODE_API_KEY_INVALID = "50111"
RET_CODE_INVALID_SIGNATURE = "50113"
RET_CODE_CANCEL_FAILED_BECAUSE_ORDER_NOT_EXISTS = "51603"
RET_CODE_ORDER_ALREADY_CANCELLED = "51401"



from decimal import Decimal

from pydantic import ConfigDict, Field, SecretStr

from hummingbot.client.config.config_data_types import BaseConnectorConfigMap
from hummingbot.core.data_type.trade_fee import TradeFeeSchema

# Maker rebates(-0.02%) are paid out continuously on each trade directly to the trading wallet.(https://derive.gitbook.io/derive-docs/trading/fees)
DEFAULT_FEES = TradeFeeSchema(
    maker_percent_fee_decimal=Decimal("0.01"),
    taker_percent_fee_decimal=Decimal("0.03"),
    buy_percent_fee_deducted_from_returns=True
)

CENTRALIZED = False

EXAMPLE_PAIR = "OP-USDC"

BROKER_ID = "HBOT"


class DeriveConfigMap(BaseConnectorConfigMap):
    connector: str = "derive"
    derive_api_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Derive Wallet address",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    derive_api_secret: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your wallet private key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    sub_id: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Subaccount Id",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    account_type: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Derive Account Type (trader/market_maker)",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        },
    )


KEYS = DeriveConfigMap.model_construct()

OTHER_DOMAINS = ["derive_testnet"]
OTHER_DOMAINS_PARAMETER = {"derive_testnet": "derive_testnet"}
OTHER_DOMAINS_EXAMPLE_PAIR = {"derive_testnet": "BTC-USD"}
OTHER_DOMAINS_DEFAULT_FEES = {"derive_testnet": [0, 0.025]}


class DeriveTestnetConfigMap(BaseConnectorConfigMap):
    connector: str = "derive_testnet"
    derive_testnet_api_key: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Derive Wallet address",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    derive_testnet_api_secret: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your wallet private key",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    sub_id: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Subaccount Id",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    account_type: SecretStr = Field(
        default=...,
        json_schema_extra={
            "prompt": "Enter your Derive Account Type (trader/market_maker)",
            "is_secure": True,
            "is_connect_key": True,
            "prompt_on_new": True,
        }
    )
    model_config = ConfigDict(title="derive")


OTHER_DOMAINS_KEYS = {"derive_testnet": DeriveTestnetConfigMap.model_construct()}



import asyncio
import logging
import math
from decimal import Decimal
from typing import Dict, List, Optional, Union

from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.core.data_type.common import OrderType, PositionAction, PriceType, TradeType
from hummingbot.core.data_type.order_candidate import OrderCandidate, PerpetualOrderCandidate
from hummingbot.core.event.events import (
    BuyOrderCreatedEvent,
    MarketOrderFailureEvent,
    OrderFilledEvent,
    SellOrderCreatedEvent,
)
from hummingbot.logger import HummingbotLogger
from hummingbot.strategy.script_strategy_base import ScriptStrategyBase
from hummingbot.strategy_v2.executors.dca_executor.data_types import DCAExecutorConfig, DCAMode
from hummingbot.strategy_v2.executors.executor_base import ExecutorBase
from hummingbot.strategy_v2.models.base import RunnableStatus
from hummingbot.strategy_v2.models.executors import CloseType, TrackedOrder


class DCAExecutor(ExecutorBase):
    _logger = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, strategy: ScriptStrategyBase, config: DCAExecutorConfig, update_interval: float = 1.0,
                 max_retries: int = 15):
        # validate amounts and prices
        if len(config.amounts_quote) != len(config.prices):
            raise ValueError("Amounts and prices lists must have the same length")

        # Initialize super class
        super().__init__(strategy=strategy, connectors=[config.connector_name], config=config, update_interval=update_interval)
        self.config: DCAExecutorConfig = config

        # validate amounts with exchange trading rules
        if self.is_any_amount_lower_than_min_order_size():
            self.close_execution_by(CloseType.FAILED)
            trading_rules = self.get_trading_rules(connector_name=config.connector_name, trading_pair=config.trading_pair)
            self.logger().error("Please increase the amount of the order:"
                                f"- Current amounts quote: {config.amounts_quote} | Min notional size: {trading_rules.min_notional_size}"
                                f"- Current amounts base: {[amount / price for amount, price in zip(config.amounts_quote, config.prices)]} | Min order size: {trading_rules.min_order_size}")
        # set default bounds
        self.n_levels = len(config.amounts_quote)
        if self.config.mode == DCAMode.TAKER and not self.config.activation_bounds:
            self.config.activation_bounds = [Decimal("0.0001"), Decimal("0.005")]  # 0.01% and 0.5%

        # executors tracking
        self._open_orders: List[TrackedOrder] = []
        self._close_orders: List[TrackedOrder] = []  # for now will be just one order but we can have multiple
        self._failed_orders: List[TrackedOrder] = []
        self._trailing_stop_trigger_pct: Optional[Decimal] = None

        # used to track the total amount filled that is updated by the event in case that the InFlightOrder is
        # not available
        self._total_executed_amount_backup: Decimal = Decimal("0")

        # add retries
        self._current_retries = 0
        self._max_retries = max_retries

    @property
    def active_open_orders(self) -> List[TrackedOrder]:
        return self._open_orders

    @property
    def active_close_orders(self) -> List[TrackedOrder]:
        return self._close_orders

    @property
    def open_order_type(self) -> OrderType:
        return OrderType.LIMIT if self.config.mode == DCAMode.MAKER else OrderType.MARKET

    @property
    def close_order_type(self) -> OrderType:
        return OrderType.MARKET

    @property
    def open_filled_amount(self) -> Decimal:
        return sum([order.executed_amount_base for order in self.active_open_orders])

    @property
    def open_filled_amount_quote(self) -> Decimal:
        return self.open_filled_amount * self.current_position_average_price

    @property
    def close_filled_amount(self) -> Decimal:
        return sum([order.executed_amount_base for order in self.active_close_orders])

    @property
    def close_filled_amount_quote(self) -> Decimal:
        return self.close_filled_amount * self.close_price

    @property
    def filled_amount(self) -> Decimal:
        return self.open_filled_amount + self.close_filled_amount

    @property
    def filled_amount_quote(self) -> Decimal:
        return self.open_filled_amount_quote + self.close_filled_amount_quote

    @property
    def max_amount_quote(self) -> Decimal:
        return sum(self.config.amounts_quote)

    @property
    def unrealized_pnl_when_last_order_filled(self) -> Decimal:
        last_order_price = self.max_price if self.config.side == TradeType.SELL else self.min_price
        distance_from_last_order_to_break_even = abs(last_order_price - self.target_position_average_price) / \
            self.target_position_average_price
        return self.max_amount_quote * distance_from_last_order_to_break_even

    @property
    def end_time(self):
        if not self.config.time_limit:
            return None
        return self.config.timestamp + self.config.time_limit

    @property
    def is_expired(self):
        return self.end_time and self.end_time <= self._strategy.current_timestamp

    @property
    def min_price(self) -> Decimal:
        return min(self.config.prices)

    @property
    def max_price(self) -> Decimal:
        return max(self.config.prices)

    @property
    def max_loss_quote(self) -> Decimal:
        # TODO: refactor the ExecutorBase class to handle max loss in pct and quote asset since some strategies like
        #  arbitrage and XEMM will need a more complex calculation
        sl = self.config.stop_loss if self.config.stop_loss else Decimal("0")
        return self.max_amount_quote * sl

    @property
    def current_market_price(self):
        """
        This method is responsible for getting the current market price to be used as a reference for control barriers
        """
        price_type = PriceType.BestBid if self.config.side == TradeType.BUY else PriceType.BestAsk
        return self.get_price(self.config.connector_name, self.config.trading_pair, price_type=price_type)

    @property
    def close_price(self):
        """
        This method is responsible for getting the close price, if the executor is active, it will return the current
        market price, otherwise it will return the average price of the closed orders
        """
        if self.status == RunnableStatus.TERMINATED and len(self._close_orders) > 0:
            # for now we will consider just one close order, in the future we can have multiple close orders
            return self._close_orders[0].average_executed_price
        else:
            return self.current_market_price

    @property
    def current_position_average_price(self) -> Decimal:
        return sum([order.average_executed_price * order.executed_amount_base for order in self._open_orders]) / \
            self.open_filled_amount if self._open_orders and self.open_filled_amount > Decimal("0") else Decimal("0")

    @property
    def target_position_average_price(self) -> Decimal:
        return sum([price * amount for price, amount in
                    zip(self.config.prices, self.config.amounts_quote)]) / self.max_amount_quote

    @property
    def trade_pnl_pct(self):
        """
        This method is responsible for calculating the trade pnl (Pure pnl without fees)
        """
        if self.current_position_average_price != Decimal("0"):
            if self.config.side == TradeType.BUY:
                return (self.close_price - self.current_position_average_price) / self.current_position_average_price
            else:
                return (self.current_position_average_price - self.close_price) / self.current_position_average_price
        else:
            return Decimal("0")

    @property
    def trade_pnl_quote(self) -> Decimal:
        """
        This method is responsible for calculating the trade pnl in quote asset
        """
        return self.trade_pnl_pct * self.open_filled_amount_quote

    def is_any_amount_lower_than_min_order_size(self):
        """
        This method is responsible for checking if any amount is lower than the minimum order size
        """
        notional_size_check = any([amount < self.connectors[self.config.connector_name].trading_rules[self.config.trading_pair].min_notional_size for amount in self.config.amounts_quote])
        base_amount_size_check = any([amount / price < self.connectors[self.config.connector_name].trading_rules[self.config.trading_pair].min_order_size for amount, price in zip(self.config.amounts_quote, self.config.prices)])
        return notional_size_check or base_amount_size_check

    def get_net_pnl_quote(self) -> Decimal:
        """
        This method is responsible for calculating the net pnl in quote asset
        """
        return self.trade_pnl_quote - self.cum_fees_quote

    def get_net_pnl_pct(self) -> Decimal:
        """
        This method is responsible for calculating the net pnl percentage
        """
        return self.net_pnl_quote / self.open_filled_amount_quote if self.open_filled_amount_quote > Decimal("0") else Decimal("0")

    def get_cum_fees_quote(self) -> Decimal:
        """
        This method is responsible for calculating the cumulative fees in quote asset
        """
        all_orders = self._open_orders + self._close_orders
        return sum([order.cum_fees_quote for order in all_orders])

    async def on_start(self):
        await super().on_start()
        if self.is_expired:
            self.close_execution_by(CloseType.EXPIRED)

    @property
    def all_open_orders_executed(self) -> bool:
        """
        This method is responsible for checking if all orders are completed
        """
        return all([order.is_done for order in self._open_orders]) and len(self._open_orders) == self.n_levels

    def validate_sufficient_balance(self):
        """
        This method is responsible for checking the budget
        """
        order_candidates = []
        for amount_quote, price in zip(self.config.amounts_quote, self.config.prices):
            amount_base = amount_quote / price
            is_maker = self.config.mode == DCAMode.MAKER
            if self.is_perpetual_connector(self.config.connector_name):
                order_candidate = PerpetualOrderCandidate(
                    trading_pair=self.config.trading_pair,
                    is_maker=is_maker,
                    order_type=self.open_order_type,
                    order_side=self.config.side,
                    amount=amount_base,
                    price=price,
                    leverage=Decimal(self.config.leverage),
                )
            else:
                order_candidate = OrderCandidate(
                    trading_pair=self.config.trading_pair,
                    is_maker=is_maker,
                    order_type=self.open_order_type,
                    order_side=self.config.side,
                    amount=amount_base,
                    price=price,
                )
            order_candidates.append(order_candidate)
        adjusted_order_candidates = self.adjust_order_candidates(self.config.connector_name, order_candidates)
        if any([order_candidate.amount == Decimal("0") for order_candidate in adjusted_order_candidates]):
            self.close_execution_by(CloseType.INSUFFICIENT_BALANCE)
            self.logger().error("Not enough budget to create DCA.")

    async def control_task(self):
        """
        This task is responsible for creating and closing position executors
        """
        if self.status == RunnableStatus.RUNNING:
            self.control_open_order_process()
            self.control_barriers()
        elif self.status == RunnableStatus.SHUTTING_DOWN:
            await self.control_shutdown_process()
        self.evaluate_max_retries()

    def control_open_order_process(self):
        """
        This method is responsible for controlling the opening process
        """
        next_level = len(self._open_orders)
        if next_level < self.n_levels:
            close_price = self.get_price(connector_name=self.config.connector_name,
                                         trading_pair=self.config.trading_pair)
            order_price = self.config.prices[next_level]
            if self._is_within_activation_bounds(order_price, close_price) and not self.is_expired:
                self.create_dca_order(level=next_level)

    def create_dca_order(self, level: int):
        """
        This method is responsible for creating a new DCA order
        """
        price = self.config.prices[level]
        amount = self.config.amounts_quote[level] / price
        order_id = self.place_order(connector_name=self.config.connector_name,
                                    trading_pair=self.config.trading_pair, order_type=self.open_order_type,
                                    side=self.config.side, amount=amount, price=price,
                                    position_action=PositionAction.OPEN)
        if order_id:
            self._open_orders.append(TrackedOrder(order_id=order_id))

    def control_barriers(self):
        """
        This method is responsible for controlling the active executors
        """
        self.control_stop_loss()
        self.control_trailing_stop()
        self.control_take_profit()
        self.control_time_limit()

    def control_time_limit(self):
        if self.is_expired:
            self.close_type = CloseType.TIME_LIMIT
            self.place_close_order_and_cancel_open_orders()

    def control_stop_loss(self):
        """
        This method is responsible for controlling the stop loss. In order to trigger the stop loss all the orders must
        be completed and the net pnl must be lower than the stop loss. If it's maker mode, the stop loss will be
        triggered if the net pnl is lower than the stop loss and all the orders were executed, otherwise the stop loss
        will be triggered if the net pnl is lower than the stop loss.
        """
        if self.config.stop_loss:
            if self.config.mode == DCAMode.MAKER:
                if self.all_open_orders_executed and self.net_pnl_pct <= -self.config.stop_loss:
                    self.close_type = CloseType.STOP_LOSS
                    self.place_close_order_and_cancel_open_orders()
            else:
                if self.net_pnl_quote <= -self.max_loss_quote:
                    self.close_type = CloseType.STOP_LOSS
                    self.place_close_order_and_cancel_open_orders()

    def control_trailing_stop(self):
        """
        This method is responsible for controlling the trailing stop. In order to activated the trailing stop the net
        pnl must be higher than the activation price delta. Once the trailing stop is activated, the trailing stop trigger
        will be the activation price delta minus the trailing delta and the stop loss will be triggered if the net pnl
        is lower than the trailing stop trigger. the value of the trailing stop trigger will be updated if the net pnl
        minus the trailing delta is higher than the current value of the trailing stop trigger.
        """
        if self.config.trailing_stop:
            net_pnl_pct = self.get_net_pnl_pct()
            if not self._trailing_stop_trigger_pct:
                if net_pnl_pct > self.config.trailing_stop.activation_price:
                    self._trailing_stop_trigger_pct = net_pnl_pct - self.config.trailing_stop.trailing_delta
            else:
                if net_pnl_pct < self._trailing_stop_trigger_pct:
                    self.close_type = CloseType.TRAILING_STOP
                    self.place_close_order_and_cancel_open_orders()
                if net_pnl_pct - self.config.trailing_stop.trailing_delta > self._trailing_stop_trigger_pct:
                    self._trailing_stop_trigger_pct = net_pnl_pct - self.config.trailing_stop.trailing_delta

    def control_take_profit(self):
        """
        This method is responsible for controlling the take profit. In order to trigger the take profit all the orders must
        be completed and the net pnl must be higher than the take profit. Take profit order is limit type only, if you
        want to use market order, you can use trailing stop instead.
        """
        if self.config.take_profit:
            if self.net_pnl_pct > self.config.take_profit:
                self.close_type = CloseType.TAKE_PROFIT
                self.place_close_order_and_cancel_open_orders()

    def early_stop(self, keep_position: bool = False):
        """
        This method allows strategy to stop the executor early.
        """
        if keep_position:
            self.close_type = CloseType.POSITION_HOLD
            self.cancel_open_orders()
            self.stop()
        else:
            self.close_type = CloseType.EARLY_STOP
            self.place_close_order_and_cancel_open_orders()

    def place_close_order_and_cancel_open_orders(self, price: Decimal = Decimal("NaN")):
        """
        This method is responsible for placing the close order
        """
        self.cancel_open_orders()
        self.place_close_order(price)
        self._status = RunnableStatus.SHUTTING_DOWN
        self.close_timestamp = self._strategy.current_timestamp

    def close_execution_by(self, close_type):
        self.close_type = close_type
        self.close_timestamp = self._strategy.current_timestamp
        self.stop()

    def place_close_order(self, price):
        delta_amount_to_close = self.open_filled_amount - self.close_filled_amount
        min_order_size = self.connectors[self.config.connector_name].trading_rules[self.config.trading_pair].min_order_size
        if delta_amount_to_close >= min_order_size:
            order_id = self.place_order(
                connector_name=self.config.connector_name,
                trading_pair=self.config.trading_pair,
                order_type=OrderType.MARKET,
                amount=delta_amount_to_close,
                price=price,
                side=TradeType.SELL if self.config.side == TradeType.BUY else TradeType.BUY,
                position_action=PositionAction.CLOSE,
            )
            self._close_orders.append(TrackedOrder(order_id=order_id))

    def cancel_open_orders(self):
        for tracked_order in self._open_orders:
            if tracked_order.order and tracked_order.order.is_open:
                self._strategy.cancel(connector_name=self.config.connector_name, trading_pair=self.config.trading_pair,
                                      order_id=tracked_order.order_id)

    def _is_within_activation_bounds(self, order_price: Decimal, close_price: Decimal) -> bool:
        """
        This method is responsible for checking if the order is within the activation bounds
        """
        activation_bounds = self.config.activation_bounds
        if self.config.mode == DCAMode.MAKER:
            if activation_bounds:
                if self.config.side == TradeType.BUY:
                    return order_price > close_price * (1 - activation_bounds[0])
                else:
                    return order_price < close_price * (1 + activation_bounds[0])
            else:
                return True
        elif self.config.mode == DCAMode.TAKER:
            # Taker mode requires activation bounds for safety. Default to 0.01% and 0.5% if not provided.
            if self.config.side == TradeType.BUY:
                min_price_to_buy = order_price * (1 - activation_bounds[0])
                max_price_to_buy = order_price * (1 + activation_bounds[1])
                return min_price_to_buy < close_price < max_price_to_buy
            else:
                min_price_to_sell = order_price * (1 - activation_bounds[1])
                max_price_to_sell = order_price * (1 + activation_bounds[0])
                return min_price_to_sell < close_price < max_price_to_sell

    async def control_shutdown_process(self):
        """
        This method is responsible for shutting down the process, ensuring that all orders are completed.
        """
        if math.isclose(self.open_filled_amount, self.close_filled_amount):
            self.close_execution_by(self.close_type)
        elif len(self.active_close_orders) > 0:
            connector = self.connectors[self.config.connector_name]
            await connector._update_orders_with_error_handler(
                orders=[order.order for order in self.active_close_orders if order.order],
                error_handler=connector._handle_update_error_for_active_order
            )
            for order in self.active_close_orders:
                self.update_tracked_orders_with_order_id(order.order_id)
                if order.order and order.order.is_done and order.executed_amount_base == Decimal("0"):
                    self.logger().error(
                        f"Close order {order.order_id} is done, might be an error with this update. Cancelling the order and placing it again.")
                    self._strategy.cancel(connector_name=self.config.connector_name, trading_pair=self.config.trading_pair,
                                          order_id=order.order_id)
                    self._close_orders.remove(order)
                    self._failed_orders.append(order)
        else:
            self.logger().info(
                f"Open amount: {self.open_filled_amount}, Close amount: {self.close_filled_amount}, Back up filled amount {self._total_executed_amount_backup}")
            self.place_close_order_and_cancel_open_orders()
            self._current_retries += 1
        await asyncio.sleep(5.0)

    def update_tracked_orders_with_order_id(self, order_id: str):
        all_orders = self._open_orders + self._close_orders
        active_order = next((order for order in all_orders if order.order_id == order_id), None)
        if active_order:
            in_flight_order = self.get_in_flight_order(self.config.connector_name, order_id)
            if in_flight_order:
                active_order.order = in_flight_order

    def process_order_created_event(self,
                                    event_tag: int,
                                    market: ConnectorBase,
                                    event: Union[BuyOrderCreatedEvent, SellOrderCreatedEvent]):
        """
        This method is responsible for processing the order created event. Here we will add the InFlightOrder to the
        active orders list.
        """
        self.update_tracked_orders_with_order_id(event.order_id)

    def process_order_failed_event(self,
                                   event_tag: int,
                                   market: ConnectorBase,
                                   event: MarketOrderFailureEvent):
        """
        This method is responsible for processing the order failed event. Here we will add the InFlightOrder to the
        failed orders list.
        """
        open_order = next((order for order in self._open_orders if order.order_id == event.order_id), None)
        if open_order:
            self._failed_orders.append(open_order)
            self._open_orders.remove(open_order)
            self.logger().error(f"Order {event.order_id} failed.")
        close_order = next((order for order in self._close_orders if order.order_id == event.order_id), None)
        if close_order:
            self._failed_orders.append(close_order)
            self._close_orders.remove(close_order)
            self.logger().error(f"Order {event.order_id} failed.")
            self._current_retries += 1

    def evaluate_max_retries(self):
        """
        This method is responsible for evaluating the max retries. If the max retries is reached, the executor will be
        stopped.
        """
        if self._current_retries >= self._max_retries:
            self.close_execution_by(CloseType.FAILED)
            self.logger().error("Max retries reached. Stopping DCA executor.")

    def process_order_filled_event(self, event_tag: int, market: ConnectorBase, event: OrderFilledEvent):
        """
        This method is responsible for processing the order filled event. Here we will update the value of
        _total_executed_amount_backup, that can be used if the InFlightOrder
        is not available.
        """
        if event.order_id in [order.order_id for order in self._open_orders]:
            self._total_executed_amount_backup += event.amount
        self.update_tracked_orders_with_order_id(event.order_id)

    def get_custom_info(self) -> Dict:
        return {
            "side": self.config.side,
            "current_position_average_price": self.current_position_average_price,
            "target_position_average_price": self.target_position_average_price,
            "filled_amount": self.filled_amount,
            "filled_amount_quote": self.filled_amount_quote,
            "max_amount_quote": self.max_amount_quote,
            "min_price": self.min_price,
            "max_price": self.max_price,
            "max_loss_quote": self.max_loss_quote,
            "current_market_price": self.current_market_price,
            "close_price": self.close_price,
            "close_type": self.close_type,
            "close_timestamp": self.close_timestamp,
            "n_levels": self.n_levels,
            "trailing_stop_trigger_pct": self._trailing_stop_trigger_pct,
            "total_executed_amount_backup": self._total_executed_amount_backup,
            "current_retries": self._current_retries,
            "max_retries": self._max_retries,
            "level_id": self.config.level_id,
            "order_ids": [order.order_id for order in self._open_orders + self._close_orders],
        }



from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

EXCHANGE_NAME = "gate_io_perpetual"
DEFAULT_DOMAIN = ""
HBOT_BROKER_ID = "hummingbot"
HBOT_ORDER_ID = "t-HBOT"
MAX_ID_LEN = 30

REST_URL = "https://api.gateio.ws/api/v4"
REST_URL_AUTH = "/api/v4"
WS_URL = "wss://fx-ws.gateio.ws/v4/ws/usdt"

# Public API v4 Endpoints
EXCHANGE_INFO_URL = "futures/usdt/contracts"
TICKER_PATH_URL = "futures/usdt/tickers"
ORDER_BOOK_PATH_URL = "futures/usdt/order_book"
MY_TRADES_PATH_URL = "futures/usdt/my_trades"
MARK_PRICE_URL = "futures/usdt/contracts/{id}"
NETWORK_CHECK_PATH_URL = "futures/usdt/contracts/BTC_USDT"
FUNDING_RATE_TIME_PATH_URL = "futures/usdt/funding_rate"

ORDER_CREATE_PATH_URL = "futures/usdt/orders"
ORDER_DELETE_PATH_URL = "futures/usdt/orders/{id}"
USER_BALANCES_PATH_URL = "futures/usdt/accounts"
POSITION_INFORMATION_URL = "futures/usdt/positions"
ORDER_STATUS_PATH_URL = "futures/usdt/orders/{id}"
USER_ORDERS_PATH_URL = "futures/usdt/orders"
SET_POSITION_MODE_URL = "futures/usdt/dual_mode"
ONEWAY_SET_LEVERAGE_PATH_URL = "futures/usdt/positions/{contract}/leverage"
HEDGE_SET_LEVERAGE_PATH_URL = "futures/usdt/dual_comp/positions/{contract}/leverage"

TICKER_ENDPOINT_NAME = "futures.tickers"
TRADES_ENDPOINT_NAME = "futures.trades"
ORDER_SNAPSHOT_ENDPOINT_NAME = "futures.order_book"
ORDERS_UPDATE_ENDPOINT_NAME = "futures.order_book_update"
USER_TRADES_ENDPOINT_NAME = "futures.usertrades"
USER_ORDERS_ENDPOINT_NAME = "futures.orders"
USER_BALANCE_ENDPOINT_NAME = "futures.balances"
USER_POSITIONS_ENDPOINT_NAME = "futures.positions"
PONG_CHANNEL_NAME = "futures.pong"

# Timeouts
MESSAGE_TIMEOUT = 30.0
PING_TIMEOUT = 10.0
API_CALL_TIMEOUT = 10.0
API_MAX_RETRIES = 4

# Funding Settlement Time Span
FUNDING_SETTLEMENT_DURATION = (0, 30)  # seconds before snapshot, seconds after snapshot

# Intervals
# Only used when nothing is received from WS
SHORT_POLL_INTERVAL = 5.0
# 45 seconds should be fine since we get trades, orders and balances via WS
LONG_POLL_INTERVAL = 45.0
# One minute should be fine since we get trades, orders and balances via WS
UPDATE_ORDER_STATUS_INTERVAL = 60.0
# 10 minute interval to update trading rules, these would likely never change whilst running.
INTERVAL_TRADING_RULES = 600

PUBLIC_URL_POINTS_LIMIT_ID = "PublicPoints"
PRIVATE_URL_POINTS_LIMIT_ID = "PrivatePoints"  # includes place-orders
CANCEL_ORDERS_LIMITS_ID = "CancelOrders"
ORDER_DELETE_LIMIT_ID = "OrderDelete"
ORDER_STATUS_LIMIT_ID = "OrderStatus"
RATE_LIMITS = [
    RateLimit(limit_id=PUBLIC_URL_POINTS_LIMIT_ID, limit=300, time_interval=1),
    RateLimit(limit_id=PRIVATE_URL_POINTS_LIMIT_ID, limit=400, time_interval=1),
    RateLimit(limit_id=CANCEL_ORDERS_LIMITS_ID, limit=400, time_interval=1),
    RateLimit(limit_id=NETWORK_CHECK_PATH_URL, limit=300, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PUBLIC_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=EXCHANGE_INFO_URL, limit=300, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PUBLIC_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=ORDER_CREATE_PATH_URL, limit=100, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PRIVATE_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=ORDER_DELETE_LIMIT_ID, limit=400, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(CANCEL_ORDERS_LIMITS_ID)]),
    RateLimit(limit_id=USER_BALANCES_PATH_URL, limit=400, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PRIVATE_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=SET_POSITION_MODE_URL, limit=400, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PRIVATE_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=POSITION_INFORMATION_URL, limit=400, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PRIVATE_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=ORDER_STATUS_LIMIT_ID, limit=400, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PRIVATE_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=ONEWAY_SET_LEVERAGE_PATH_URL, limit=400, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PRIVATE_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=HEDGE_SET_LEVERAGE_PATH_URL, limit=400, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PRIVATE_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=USER_ORDERS_PATH_URL, limit=400, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PRIVATE_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=TICKER_PATH_URL, limit=300, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PUBLIC_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=MARK_PRICE_URL, limit=300, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PUBLIC_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=FUNDING_RATE_TIME_PATH_URL, limit=300, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PUBLIC_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=ORDER_BOOK_PATH_URL, limit=300, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PUBLIC_URL_POINTS_LIMIT_ID)]),
    RateLimit(limit_id=MY_TRADES_PATH_URL, limit=400, time_interval=1,
              linked_limits=[LinkedLimitWeightPair(PRIVATE_URL_POINTS_LIMIT_ID)]),
]



import asyncio
import hashlib
import time
from copy import deepcopy
from decimal import Decimal
from typing import Any, AsyncIterable, Dict, List, Optional, Tuple

from bidict import bidict

from hummingbot.connector.constants import SECOND, s_decimal_NaN
from hummingbot.connector.derivative.derive_perpetual import (
    derive_perpetual_constants as CONSTANTS,
    derive_perpetual_web_utils as web_utils,
)
from hummingbot.connector.derivative.derive_perpetual.derive_perpetual_api_order_book_data_source import (
    DerivePerpetualAPIOrderBookDataSource,
)
from hummingbot.connector.derivative.derive_perpetual.derive_perpetual_api_user_stream_data_source import (
    DerivePerpetualAPIUserStreamDataSource,
)
from hummingbot.connector.derivative.derive_perpetual.derive_perpetual_auth import DerivePerpetualAuth
from hummingbot.connector.derivative.position import Position
from hummingbot.connector.perpetual_derivative_py_base import PerpetualDerivativePyBase
from hummingbot.connector.trading_rule import TradingRule
from hummingbot.connector.utils import combine_to_hb_trading_pair, get_new_client_order_id
from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.common import OrderType, PositionAction, PositionMode, PositionSide, TradeType
from hummingbot.core.data_type.in_flight_order import InFlightOrder, OrderUpdate, TradeUpdate
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.data_type.trade_fee import TokenAmount, TradeFeeBase
from hummingbot.core.data_type.user_stream_tracker_data_source import UserStreamTrackerDataSource
from hummingbot.core.utils.async_utils import safe_ensure_future, safe_gather
from hummingbot.core.utils.estimate_fee import build_trade_fee
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory


class DerivePerpetualDerivative(PerpetualDerivativePyBase):
    UPDATE_ORDER_STATUS_MIN_INTERVAL = 10.0
    web_utils = web_utils

    SHORT_POLL_INTERVAL = 5.0
    LONG_POLL_INTERVAL = 120.0

    def __init__(
            self,
            balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
            rate_limits_share_pct: Decimal = Decimal("100"),
            derive_perpetual_api_secret: str = None,
            sub_id: int = None,
            account_type: str = None,
            derive_perpetual_api_key: str = None,
            trading_pairs: Optional[List[str]] = None,
            trading_required: bool = True,
            domain: str = CONSTANTS.DEFAULT_DOMAIN,
    ):
        self.derive_perpetual_api_key = derive_perpetual_api_key
        self.derive_perpetual_secret_key = derive_perpetual_api_secret
        self._sub_id = sub_id
        self._trading_required = trading_required
        self._trading_pairs = trading_pairs
        self._domain = domain
        self._account_type = account_type
        self._position_mode = None
        self._last_trade_history_timestamp = None
        self._last_trades_poll_timestamp = 1.0
        self._instrument_ticker = []
        self.real_time_balance_update = False
        super().__init__(balance_asset_limit, rate_limits_share_pct)

    @property
    def name(self) -> str:
        # Note: domain here refers to the entire exchange name. i.e. derive_perpetual_ or derive_perpetual_testnet
        return self._domain

    @staticmethod
    def derive_perpetual_order_type(order_type: OrderType) -> str:
        return order_type.name.lower()

    @property
    def authenticator(self) -> DerivePerpetualAuth:
        return DerivePerpetualAuth(
            self.derive_perpetual_api_key,
            self.derive_perpetual_secret_key,
            self._sub_id,
            self._trading_required,
            self._domain)

    @property
    def rate_limits_rules(self) -> List[RateLimit]:
        return CONSTANTS.RATE_LIMITS

    @property
    def domain(self) -> str:
        return self._domain

    @property
    def client_order_id_max_length(self) -> int:
        return CONSTANTS.MAX_ORDER_ID_LEN

    @property
    def client_order_id_prefix(self) -> str:
        return CONSTANTS.BROKER_ID

    @property
    def trading_rules_request_path(self) -> str:
        return CONSTANTS.EXCHANGE_INFO_PATH_URL

    @property
    def trading_pairs_request_path(self) -> str:
        return CONSTANTS.EXCHANGE_INFO_PATH_URL

    @property
    def trading_currencies_request_path(self) -> str:
        return CONSTANTS.EXCHANGE_CURRENCIES_PATH_URL

    @property
    def check_network_request_path(self) -> str:
        return CONSTANTS.PING_PATH_URL

    @property
    def trading_pairs(self):
        return self._trading_pairs

    @property
    def is_cancel_request_in_exchange_synchronous(self) -> bool:
        return True

    @property
    def is_trading_required(self) -> bool:
        return self._trading_required

    @property
    def funding_fee_poll_interval(self) -> int:
        return 120

    async def _make_network_check_request(self):
        await self._api_get(path_url=self.check_network_request_path)

    def supported_order_types(self) -> List[OrderType]:
        """
        :return a list of OrderType supported by this connector
        """
        return [OrderType.LIMIT, OrderType.LIMIT_MAKER, OrderType.MARKET]

    def supported_position_modes(self):
        """
        This method needs to be overridden to provide the accurate information depending on the exchange.
        """
        return [PositionMode.ONEWAY]

    def get_buy_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules[trading_pair]
        return trading_rule.buy_order_collateral_token

    def get_sell_collateral_token(self, trading_pair: str) -> str:
        trading_rule: TradingRule = self._trading_rules[trading_pair]
        return trading_rule.sell_order_collateral_token

    async def _make_trading_pairs_request(self) -> Any:
        payload = {
            "expired": True,
            "instrument_type": "perp",
            "page": 1,
            "page_size": 1000,
        }

        exchange_info = await self._api_post(path_url=self.trading_currencies_request_path, data=payload)
        info = exchange_info["result"]["instruments"]
        self._instrument_ticker = info
        return info

    async def _make_trading_rules_request(self) -> Any:
        payload = {
            "expired": False,
            "instrument_type": "perp",
            "page": 1,
            "page_size": 1000,
        }
        exchange_info = await self._api_post(path_url=self.trading_pairs_request_path, data=(payload))
        info: List[Dict[str, Any]] = exchange_info["result"]
        return info

    async def get_all_pairs_prices(self) -> Dict[str, Any]:
        res = []
        tasks = []
        if len(self._instrument_ticker) == 0:
            await self._make_trading_pairs_request()
        for token in self._instrument_ticker:
            payload = {"instrument_name": token["instrument_name"]}
            tasks.append(self._api_post(path_url=CONSTANTS.TICKER_PRICE_CHANGE_PATH_URL, data=payload))
        results = await safe_gather(*tasks, return_exceptions=True)
        for result in results:
            pair_price_data = result["result"]

            data = {
                "symbol": {
                    "instrument_name": pair_price_data["instrument_name"],
                    "best_bid": pair_price_data["best_bid_price"],
                    "best_ask": pair_price_data["best_ask_price"],
                }
            }
            res.append(data)
        return res

    def _is_request_exception_related_to_time_synchronizer(self, request_exception: Exception):
        return False

    def _initialize_trading_pair_symbols_from_exchange_info(self, exchange_info: List):
        mapping = bidict()

        for _info in filter(web_utils.is_exchange_information_valid, exchange_info):
            ex_name = _info["instrument_name"]

            base, _quote = ex_name.split("-")
            _quote = "USDC"
            trading_pair = combine_to_hb_trading_pair(base, _quote)
            mapping[ex_name] = trading_pair
        self._set_trading_pair_symbol_map(mapping)

    async def _update_trading_rules(self):
        exchange_info = await self._make_trading_pairs_request()
        trading_rules_list = await self._format_trading_rules(exchange_info)
        self._trading_rules.clear()
        for trading_rule in trading_rules_list:
            self._trading_rules[trading_rule.trading_pair] = trading_rule
        self._initialize_trading_pair_symbols_from_exchange_info(exchange_info=exchange_info)

    def _create_web_assistants_factory(self) -> WebAssistantsFactory:
        return web_utils.build_api_factory(
            throttler=self._throttler,
            time_synchronizer=self._time_synchronizer,
            domain=self._domain,
            auth=self._auth)

    def _create_order_book_data_source(self) -> OrderBookTrackerDataSource:
        return DerivePerpetualAPIOrderBookDataSource(
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    def _create_user_stream_data_source(self) -> UserStreamTrackerDataSource:
        return DerivePerpetualAPIUserStreamDataSource(
            auth=self._auth,
            trading_pairs=self._trading_pairs,
            connector=self,
            api_factory=self._web_assistants_factory,
            domain=self.domain,
        )

    async def start_network(self):
        await super().start_network()
        self._rate_limits_polling_task = safe_ensure_future(self._rate_limits_polling_loop())

    def _is_order_not_found_during_status_update_error(self, status_update_exception: Exception) -> bool:
        return CONSTANTS.ORDER_NOT_EXIST_MESSAGE in str(status_update_exception)

    def _is_order_not_found_during_cancelation_error(self, cancelation_exception: Exception) -> bool:
        return CONSTANTS.UNKNOWN_ORDER_MESSAGE in str(cancelation_exception)

    def quantize_order_price(self, trading_pair: str, price: Decimal) -> Decimal:
        """
        Applies trading rule to quantize order price.
        """
        d_price = Decimal(round(float(f"{price:.5g}"), 6))
        return d_price

    def _get_fee(self,
                 base_currency: str,
                 quote_currency: str,
                 order_type: OrderType,
                 order_side: TradeType,
                 amount: Decimal,
                 price: Decimal = s_decimal_NaN,
                 is_maker: Optional[bool] = None) -> TradeFeeBase:
        is_maker = order_type is OrderType.LIMIT_MAKER
        trade_base_fee = build_trade_fee(
            exchange=self.name,
            is_maker=is_maker,
            order_side=order_side,
            order_type=order_type,
            amount=amount,
            price=price,
            base_currency=base_currency.upper(),
            quote_currency=quote_currency.upper()
        )
        return trade_base_fee

    async def _status_polling_loop_fetch_updates(self):
        await safe_gather(
            self._update_trade_history(),
            self._update_order_status(),
            self._update_balances(),
            self._update_positions(),
        )

        # === loops and sync related methods === #
    async def _rate_limits_polling_loop(self):
        """
        Updates the rate limits.
        """
        try:
            await self._update_rate_limits()
        except NotImplementedError:
            raise
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().info(
                "Unexpected error while Updating rate limits."
            )

    async def _update_rate_limits(self):
        await self._initialize_rate_limits()

    async def _initialize_rate_limits(self):
        # Update rate limits
        for r_limit_id in CONSTANTS.ENDPOINTS["limits"]["non_matching"]:
            limit_id = None
            rate_limits_copy = deepcopy(self._throttler._rate_limits)

            if self._account_type == CONSTANTS.MARKET_MAKER_ACCOUNTS_TYPE:
                limit_id = r_limit_id
                interval = SECOND
                limit = CONSTANTS.MARKET_MAKER_NON_MATCHING
            else:
                limit_id = r_limit_id
                interval = SECOND
                limit = CONSTANTS.TRADER_NON_MATCHING

            if limit_id is not None and interval is not None:
                for r_l in rate_limits_copy:
                    if r_l.limit_id == limit_id:
                        rate_limits_copy.remove(r_l)
                rate_limits_copy.append(
                    RateLimit(
                        limit_id=limit_id,
                        limit=limit,
                        time_interval=interval,
                    )
                )
            self._throttler.set_rate_limits(rate_limits_copy)

    async def _update_order_status(self):
        await self._update_orders()

    async def _update_lost_orders_status(self):
        await self._update_lost_orders()

    async def _update_trading_fees(self):
        """
        Update fees information from the derivative exchange.
        """
        pass

    async def _place_cancel(self, order_id: str, tracked_order: InFlightOrder):
        oid = await tracked_order.get_exchange_order_id()
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=tracked_order.trading_pair)
        api_params = {
            "instrument_name": symbol,
            "order_id": oid,
            "subaccount_id": int(self._sub_id)
        }
        cancel_result = await self._api_post(
            path_url=CONSTANTS.CANCEL_ORDER_URL,
            data=api_params,
            is_auth_required=True)

        if "error" in cancel_result:
            if 'Does not exist' in cancel_result['error']['message']:
                self.logger().debug(f"The order {order_id} does not exist on DerivePerpetual s. "
                                    f"No cancelation needed.")
                await self._order_tracker.process_order_not_found(order_id)
            raise IOError(f'{cancel_result["error"]["message"]}')
        if "result" in cancel_result:
            if cancel_result["result"]["order_status"] == "cancelled":
                return True
        return False

    # === Orders placing ===

    def buy(self,
            trading_pair: str,
            amount: Decimal,
            order_type=OrderType.LIMIT,
            price: Decimal = s_decimal_NaN,
            **kwargs) -> str:
        """
        Creates a promise to create a buy order using the parameters

        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price

        :return: the id assigned by the connector to the order (the client id)
        """
        order_id = get_new_client_order_id(
            is_buy=True,
            trading_pair=trading_pair,
            hbot_order_id_prefix=self.client_order_id_prefix,
            max_id_len=self.client_order_id_max_length
        )
        md5 = hashlib.md5()
        md5.update(order_id.encode('utf-8'))
        hex_order_id = f"0x{md5.hexdigest()}"
        if order_type is OrderType.MARKET:
            mid_price = self.get_mid_price(trading_pair)
            price = self.quantize_order_price(trading_pair, mid_price)

        safe_ensure_future(self._create_order(
            trade_type=TradeType.BUY,
            order_id=hex_order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price,
            **kwargs))
        return hex_order_id

    def sell(self,
             trading_pair: str,
             amount: Decimal,
             order_type: OrderType = OrderType.LIMIT,
             price: Decimal = s_decimal_NaN,
             **kwargs) -> str:
        """
        Creates a promise to create a sell order using the parameters.
        :param trading_pair: the token pair to operate with
        :param amount: the order amount
        :param order_type: the type of order to create (MARKET, LIMIT, LIMIT_MAKER)
        :param price: the order price
        :return: the id assigned by the connector to the order (the client id)
        """
        order_id = get_new_client_order_id(
            is_buy=False,
            trading_pair=trading_pair,
            hbot_order_id_prefix=self.client_order_id_prefix,
            max_id_len=self.client_order_id_max_length
        )
        md5 = hashlib.md5()
        md5.update(order_id.encode('utf-8'))
        hex_order_id = f"0x{md5.hexdigest()}"
        if order_type is OrderType.MARKET:
            mid_price = self.get_mid_price(trading_pair)
            price = self.quantize_order_price(trading_pair, mid_price)

        safe_ensure_future(self._create_order(
            trade_type=TradeType.SELL,
            order_id=hex_order_id,
            trading_pair=trading_pair,
            amount=amount,
            order_type=order_type,
            price=price,
            **kwargs))
        return hex_order_id

    async def _place_order(
            self,
            order_id: str,
            trading_pair: str,
            amount: Decimal,
            trade_type: TradeType,
            order_type: OrderType,
            price: Decimal,
            position_action: PositionAction = PositionAction.NIL,
            **kwargs,
    ) -> Tuple[str, float]:
        """
        Creates an order on the derivative exchange using the specified parameters.
        """
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        if len(self._instrument_ticker) == 0:
            await self._make_trading_pairs_request()
        instrument = next((pair for pair in self._instrument_ticker if symbol == pair["instrument_name"]), None)
        if order_type is OrderType.LIMIT and position_action == PositionAction.CLOSE:
            param_order_type = "gtc"
        elif order_type is OrderType.LIMIT_MAKER:
            param_order_type = "gtc"
        elif order_type is OrderType.MARKET:
            param_order_type = "ioc"
        else:
            param_order_type = "gtc"
        type_str = DerivePerpetualDerivative.derive_perpetual_order_type(order_type)

        price_type = "limit" if type_str == "limit_maker" or type_str == "limit" else "market"
        new_price = float(f"{price:.4g}")
        api_params = {
            "asset_address": instrument["base_asset_address"],
            "sub_id": instrument["base_asset_sub_id"],
            "limit_price": str(new_price),
            "type": "order",
            "max_fee": str(1000),
            "amount": str(amount),
            "instrument_name": symbol,
            "label": order_id,
            "is_bid": True if trade_type is TradeType.BUY else False,
            "direction": "buy" if trade_type is TradeType.BUY else "sell",
            "order_type": price_type,
            "reduce_only": False,
            "referral_code": CONSTANTS.REFERRAL_CODE,
            "mmp": False,
            "time_in_force": param_order_type,
            "recipient_id": self._sub_id,
        }

        order_result = await self._api_post(
            path_url = CONSTANTS.CREATE_ORDER_URL,
            data=api_params,
            is_auth_required=True)

        if "error" in order_result:
            if "Self-crossing disallowed" in order_result["error"]["message"]:
                self.logger().warning(f"Error submitting order: {order_result['error']['data']}")
            else:
                raise IOError(f"Error submitting order {order_id}: {order_result['error']['data']}")
        else:
            o_data = order_result['result'].get("order")
            return (str(o_data["order_id"]), o_data["creation_timestamp"] * 1e-3)

    async def _update_trade_history(self):
        orders = list(self._order_tracker.all_fillable_orders.values())
        all_fillable_orders = self._order_tracker.all_fillable_orders_by_exchange_order_id
        all_fills_response = []
        if len(orders) > 0:
            try:
                all_fills_response = await self._api_get(
                    path_url=CONSTANTS.MY_TRADES_PATH_URL,
                    params={
                        "subaccount_id": self._sub_id
                    },
                    is_auth_required=True,
                    limit_id=CONSTANTS.MY_TRADES_PATH_URL)
            except asyncio.CancelledError:
                raise
            except Exception as request_error:
                self.logger().warning(
                    f"Failed to fetch trade updates. Error: {request_error}",
                    exc_info = request_error,
                )
            for trade_fill in all_fills_response["result"]["trades"]:
                await self._process_trade_rs_event_message(order_fill=trade_fill, all_fillable_order=all_fillable_orders)

    async def _process_trade_rs_event_message(self, order_fill: Dict[str, Any], all_fillable_order):
        exchange_order_id = str(order_fill.get("order_id"))
        fillable_order = all_fillable_order.get(exchange_order_id)
        if fillable_order is not None:
            fee_asset = fillable_order.quote_asset
            position_side = PositionSide.LONG if order_fill["direction"] == 'buy' else PositionSide.SHORT
            position_action = (PositionAction.OPEN
                               if (fillable_order.trade_type is TradeType.BUY and position_side == "LONG"
                                   or fillable_order.trade_type is TradeType.SELL and position_side == "SHORT")
                               else PositionAction.CLOSE)
            fee = TradeFeeBase.new_perpetual_fee(
                fee_schema=self.trade_fee_schema(),
                position_action=position_action,
                percent_token=fee_asset,
                flat_fees=[TokenAmount(amount=Decimal(order_fill["trade_fee"]), token=fee_asset)]
            )

            trade_update = TradeUpdate(
                trade_id=str(order_fill["trade_id"]),
                client_order_id=fillable_order.client_order_id,
                exchange_order_id=str(order_fill["order_id"]),
                trading_pair=fillable_order.trading_pair,
                fee=fee,
                fill_base_amount=Decimal(order_fill["trade_amount"]),
                fill_quote_amount=Decimal(order_fill["trade_price"]) * Decimal(order_fill["trade_amount"]),
                fill_price=Decimal(order_fill["trade_price"]),
                fill_timestamp=order_fill["timestamp"] * 1e-3,
            )

            self._order_tracker.process_trade_update(trade_update)

    async def _iter_user_event_queue(self) -> AsyncIterable[Dict[str, any]]:
        while True:
            try:
                yield await self._user_stream_tracker.user_stream.get()
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().network(
                    "Unknown error. Retrying after 1 seconds.",
                    exc_info=True,
                    app_warning_msg="Could not fetch user events from DerivePerpetual. Check API key and network connection.",
                )
                await self._sleep(1.0)

    async def _user_stream_event_listener(self):
        """
        Listens to messages from _user_stream_tracker.user_stream queue.
        Traders, Orders, and Balance updates from the WS.
        """
        user_channels = [
            f"{self._sub_id}.{CONSTANTS.USER_ORDERS_ENDPOINT_NAME}",
            f"{self._sub_id}.{CONSTANTS.USEREVENT_ENDPOINT_NAME}",
            "positions",
            "collaterals",
        ]
        async for event_message in self._iter_user_event_queue():
            try:
                if isinstance(event_message, dict):
                    if "channel" in event_message:
                        channel: str = event_message.get("channel", None)
                        results = event_message.get("data", None)
                    else:
                        if "collaterals" in event_message["result"]:
                            channel = "collaterals"
                            results = event_message.get("result", {})
                        elif "positions" in event_message["result"]:
                            channel = "positions"
                            results = event_message.get("result", {})
                        else:
                            channel = "none"

                elif event_message is asyncio.CancelledError:
                    raise asyncio.CancelledError
                else:
                    raise Exception(event_message)
                if channel not in user_channels:
                    self.logger().error(
                        f"Unexpected message in user stream: {event_message}.", exc_info=True)
                    continue
                if channel == user_channels[0] and results is not None:
                    for order_msg in results:
                        self._process_order_message(order_msg)
                elif channel == user_channels[1] and results is not None:
                    for trade_msg in results:
                        await self._process_trade_message(trade_msg)
                elif channel == user_channels[2] and results is not None:
                    await self._process_update_positions(results)
                elif channel == user_channels[3] and results is not None:
                    for balance_msg in results["collaterals"]:
                        self._process_update_balances(balance_msg)
            except asyncio.CancelledError:
                raise
            except Exception:
                self.logger().error(
                    "Unexpected error in user stream listener loop.", exc_info=True)
                await self._sleep(5.0)

    async def _process_update_positions(self, results):
        for asset in results["positions"]:
            trading_pair = asset["instrument_name"]
            try:
                hb_trading_pair = await self.trading_pair_associated_to_exchange_symbol(trading_pair)
                if hb_trading_pair in self.trading_pairs:
                    position_side = PositionSide.LONG if Decimal(asset.get("amount")) > 0 else PositionSide.SHORT
                    pos_key = self._perpetual_trading.position_key(hb_trading_pair, position_side)
                    position = self._perpetual_trading.get_position(hb_trading_pair, position_side)
                    trading_rule = self._trading_rules[trading_pair]
                    amount_precision = Decimal(trading_rule.min_base_amount_increment)
                    amount = Decimal(asset.get("amount"), 0)
                    if position is not None:
                        if amount == Decimal("0"):
                            self._perpetual_trading.remove_position(pos_key)
                        else:
                            entry_price = Decimal(asset.get("index_price"))
                            unrealized_pnl = Decimal(asset.get("unrealized_pnl"))
                            position.update_position(position_side=position_side,
                                                     unrealized_pnl=unrealized_pnl,
                                                     entry_price=entry_price,
                                                     amount=Decimal(amount * amount_precision))
                    else:
                        await self._update_positions()
            except KeyError:
                continue

    def _process_update_balances(self, balance_msg):
        asset_name = balance_msg["asset_name"]
        self._account_balances[asset_name] = Decimal(balance_msg["amount"])
        self._account_available_balances[asset_name] = Decimal(balance_msg["amount"])

    async def _process_trade_message(self, trade: Dict[str, Any], client_order_id: Optional[str] = None):
        """
        Updates in-flight order and trigger order filled event for trade message received. Triggers order completed
        event if the total executed amount equals to the specified order amount.
        Example Trade:
        """
        exchange_order_id = str(trade.get("order_id", ""))
        tracked_order = self._order_tracker.all_fillable_orders_by_exchange_order_id.get(exchange_order_id)

        if tracked_order is None:
            all_orders = self._order_tracker.all_fillable_orders
            for k, v in all_orders.items():
                await v.get_exchange_order_id()
            _cli_tracked_orders = [o for o in all_orders.values() if exchange_order_id == o.exchange_order_id]
            if not _cli_tracked_orders:
                self.logger().debug(f"Ignoring trade message with id {client_order_id}: not in in_flight_orders.")
                return
            tracked_order = _cli_tracked_orders[0]
        trading_pair = tracked_order.trading_pair
        symbol = await self.trading_pair_associated_to_exchange_symbol(symbol=trade["instrument_name"])
        if symbol == trading_pair:
            fee_asset = tracked_order.quote_asset
            position_side = PositionSide.LONG if trade["direction"] == 'buy' else PositionSide.SHORT
            position_action = (PositionAction.OPEN
                               if (tracked_order.trade_type is TradeType.BUY and position_side == "LONG"
                                   or tracked_order.trade_type is TradeType.SELL and position_side == "SHORT")
                               else PositionAction.CLOSE)
            fee = TradeFeeBase.new_perpetual_fee(
                fee_schema=self.trade_fee_schema(),
                position_action=position_action,
                percent_token=fee_asset,
                flat_fees=[TokenAmount(amount=Decimal(trade["trade_fee"]), token=fee_asset)]
            )
            trade_update: TradeUpdate = TradeUpdate(
                trade_id=str(trade["trade_id"]),
                client_order_id=tracked_order.client_order_id,
                exchange_order_id=str(trade["order_id"]),
                trading_pair=tracked_order.trading_pair,
                fill_timestamp=trade["timestamp"] * 1e-3,
                fill_price=Decimal(trade["trade_price"]),
                fill_base_amount=Decimal(trade["trade_amount"]),
                fill_quote_amount=Decimal(trade["trade_price"]) * Decimal(trade["trade_amount"]),
                fee=fee,
            )
            self._order_tracker.process_trade_update(trade_update)
            await self._update_positions()

    def _process_order_message(self, order_msg: Dict[str, Any]):
        """
        Updates in-flight order and triggers cancelation or failure event if needed.

        :param order_msg: The order response from either REST or web socket API (they are of the same format)

        Example Order:
        """
        client_order_id = str(order_msg.get("label", ""))
        tracked_order = self._order_tracker.all_updatable_orders.get(client_order_id)
        if not tracked_order:
            self.logger().debug(f"Ignoring order message with id {client_order_id}: not in in_flight_orders.")
            return
        current_state = order_msg["order_status"]
        order_update: OrderUpdate = OrderUpdate(
            trading_pair=tracked_order.trading_pair,
            update_timestamp=order_msg["last_update_timestamp"] * 1e-3,
            new_state=CONSTANTS.ORDER_STATE[current_state],
            client_order_id=order_msg["label"],
            exchange_order_id=str(order_msg["order_id"]),
        )
        self._order_tracker.process_order_update(order_update=order_update)

    async def _format_trading_rules(self, exchange_info_dict: List) -> List[TradingRule]:
        """
        Queries the necessary API endpoint and initialize the TradingRule object for each trading pair being traded.

        Parameters
        ----------
        exchange_info_dict:
            Trading rules dictionary response from the derivative

        {
            "result": {
                "instruments": [
                {
                    "instrument_type": "perp",
                    "instrument_name": "OP-USDC",
                    "scheduled_activation": 1728508925,
                    "scheduled_deactivation": 9223372036854776000,
                    "is_active": true,
                    "tick_size": "0.01",
                    "minimum_amount": "0.1",
                    "maximum_amount": "1000",
                    "amount_step": "0.01",
                    "mark_price_fee_rate_cap": "0",
                    "maker_fee_rate": "0.0015",
                    "taker_fee_rate": "0.0015",
                    "base_fee": "0.1",
                    "base_currency": "OP",
                    "quote_currency": "USD",
                    "option_details": null,
                    "perp_details": null,
                    "erc20_details": {
                    "decimals": 18,
                    "underlying_erc20_address": "0x15CEcd5190A43C7798dD2058308781D0662e678E",
                    "borrow_index": "1",
                    "supply_index": "1"
                    },
                    "base_asset_address": "0xE201fCEfD4852f96810C069f66560dc25B2C7A55",
                    "base_asset_sub_id": "0",
                    "pro_rata_fraction": "0",
                    "fifo_min_allocation": "0",
                    "pro_rata_amount_step": "1"
                }
                ],
                "pagination": {
                "num_pages": 1,
                "count": 1
                }
            },
            "id": "0f9131b4-2502-4f8e-afa4-adfce67a6509"
        }
        """

        trading_pair_rules = exchange_info_dict
        retval = []
        for rule in filter(web_utils.is_exchange_information_valid, trading_pair_rules):
            if rule["instrument_type"] != "perp":
                continue
            try:
                trading_pair = await self.trading_pair_associated_to_exchange_symbol(symbol=rule["instrument_name"])
                min_order_size = rule["minimum_amount"]
                step_size = rule["amount_step"]
                tick_size = rule["tick_size"]
                collateral_token = "USDC"
                retval.append(
                    TradingRule(
                        trading_pair,
                        min_order_size=Decimal(min_order_size),
                        min_price_increment=Decimal(str(tick_size)),
                        min_base_amount_increment=Decimal(step_size),
                        buy_order_collateral_token=collateral_token,
                        sell_order_collateral_token=collateral_token,
                    )
                )
            except Exception:
                self.logger().error(f"Error parsing the trading pair rule {exchange_info_dict}. Skipping.",
                                    exc_info=True)
        return retval

    async def _update_balances(self):
        """
        Calls the REST API to update total and available balances.
        """
        local_asset_names = set(self._account_balances.keys())
        remote_asset_names = set()

        account_info = await self._api_post(
            path_url=CONSTANTS.ACCOUNTS_PATH_URL,
            data={"subaccount_id": self._sub_id},
            is_auth_required=True)
        if "error" in account_info:
            self.logger().error(f"Error fetching account balances: {account_info['error']['message']}")
            raise
        else:
            balances = account_info["result"]["collaterals"]
            for balance_entry in balances:
                asset_name = balance_entry["asset_name"]
                free_balance = Decimal(balance_entry["amount"])
                total_balance = Decimal(balance_entry["amount"])
                self._account_available_balances[asset_name] = free_balance
                self._account_balances[asset_name] = total_balance
                remote_asset_names.add(asset_name)

            asset_names_to_remove = local_asset_names.difference(remote_asset_names)
            for asset_name in asset_names_to_remove:
                del self._account_available_balances[asset_name]
                del self._account_balances[asset_name]

    async def _request_order_status(self, tracked_order: InFlightOrder) -> OrderUpdate:
        oid = await tracked_order.get_exchange_order_id()
        client_order_id = tracked_order.client_order_id
        order_update = await self._api_post(
            path_url=CONSTANTS.ORDER_STATUS_PAATH_URL,
            data={
                "subaccount_id": self._sub_id,
                "order_id": oid
            },
            is_auth_required=True)
        if "error" in order_update:
            self.logger().debug(f"Error fetching order status for {client_order_id}: {order_update['error']['message']}")
        if "result" in order_update:
            current_state = order_update["result"]["order_status"]
            _order_update: OrderUpdate = OrderUpdate(
                trading_pair=tracked_order.trading_pair,
                update_timestamp=order_update["result"]["last_update_timestamp"] * 1e-3,
                new_state=CONSTANTS.ORDER_STATE[current_state],
                client_order_id=order_update["result"]["label"] or client_order_id,
                exchange_order_id=str(order_update["result"]["order_id"]),
            )
            return _order_update

    async def _all_trade_updates_for_order(self, order: InFlightOrder) -> List[TradeUpdate]:
        exchange_order_id = str(order.exchange_order_id)
        if exchange_order_id is not None:
            trading_pair = await self.exchange_symbol_associated_to_pair(trading_pair=order.trading_pair)
            all_fills_response = await self._api_get(
                path_url=CONSTANTS.MY_TRADES_PATH_URL,
                params={
                    "instrument_name": trading_pair,
                    "order_id": exchange_order_id,
                    "subaccount_id": self._sub_id
                },
                is_auth_required=True,
                limit_id=CONSTANTS.MY_TRADES_PATH_URL)

            for trade in all_fills_response["result"]["trades"]:
                fee_asset = order.quote_asset
                if str(trade["order_id"]) == exchange_order_id:
                    position_side = PositionSide.LONG if trade["direction"] == 'buy' else PositionSide.SHORT
                    position_action = (PositionAction.OPEN
                                       if (order.trade_type is TradeType.BUY and position_side == "LONG"
                                           or order.trade_type is TradeType.SELL and position_side == "SHORT") else PositionAction.CLOSE)
                    fee = TradeFeeBase.new_perpetual_fee(
                        fee_schema=self.trade_fee_schema(),
                        position_action=position_action,
                        percent_token=fee_asset,
                        flat_fees=[TokenAmount(amount=Decimal(trade["trade_fee"]), token=fee_asset)])
                    trade_update = TradeUpdate(
                        trade_id=str(trade["trade_id"]),
                        client_order_id=order.client_order_id,
                        exchange_order_id=exchange_order_id,
                        trading_pair=trading_pair,
                        fee=fee,
                        fill_base_amount=Decimal(trade["trade_amount"]),
                        fill_quote_amount=Decimal(trade["trade_amount"]) * Decimal(trade["trade_price"]),
                        fill_price=Decimal(trade["trade_price"]),
                        fill_timestamp=trade["timestamp"] * 1e-3,
                    )
                    self._order_tracker.process_trade_update(trade_update)

    async def _get_last_traded_price(self, trading_pair: str) -> float:
        await self.trading_pair_symbol_map()
        exchange_symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        payload = {"instrument_name": exchange_symbol}
        response = await self._api_post(path_url=CONSTANTS.TICKER_PRICE_CHANGE_PATH_URL, data=payload, is_auth_required=False,
                                        limit_id=CONSTANTS.TICKER_PRICE_CHANGE_PATH_URL)

        return response["result"]["mark_price"]

    async def get_last_traded_prices(self, trading_pairs: List[str] = None) -> Dict[str, float]:
        if trading_pairs is None:
            trading_pairs = []

        symbol_map = await self.trading_pair_symbol_map()
        exchange_symbols = await asyncio.gather(*[
            self.exchange_symbol_associated_to_pair(trading_pair=pair) for pair in trading_pairs
        ])
        payloads = [{"instrument_name": symbol} for symbol in exchange_symbols]
        responses = await asyncio.gather(*[
            self._api_post(path_url=CONSTANTS.TICKER_PRICE_CHANGE_PATH_URL, data=payload)
            for payload in payloads
        ])
        last_traded_prices = {}
        for ticker in responses:
            instrument_name = ticker["result"]["instrument_name"]
            if instrument_name in symbol_map.keys():
                mapped_name = await self.trading_pair_associated_to_exchange_symbol(instrument_name)
                last_traded_prices[mapped_name] = Decimal(ticker["result"]["mark_price"])
        return last_traded_prices

    async def _update_positions(self):
        positions = await self._api_post(path_url=CONSTANTS.POSITION_INFORMATION_URL,
                                         data={"subaccount_id": self._sub_id},
                                         is_auth_required=True,
                                         limit_id=CONSTANTS.POSITION_INFORMATION_URL)
        if "result" in positions:
            data: List[dict] = positions["result"]["positions"]
            if len(data) == 0:
                return
            for position in data:
                trading_pair = position.get("instrument_name")
                try:
                    hb_trading_pair = await self.trading_pair_associated_to_exchange_symbol(trading_pair)
                except KeyError:
                    # Ignore results for which their symbols is not tracked by the connector
                    continue
                position_side = PositionSide.LONG if Decimal(position.get("amount")) > 0 else PositionSide.SHORT
                unrealized_pnl = Decimal(position.get("unrealized_pnl"))
                entry_price = Decimal(position.get("index_price"))
                amount = Decimal(position.get("amount", 0))
                leverage = position.get("leverage", 0)
                pos_key = self._perpetual_trading.position_key(hb_trading_pair, position_side)
                if amount != 0:
                    _position = Position(
                        trading_pair=hb_trading_pair,
                        position_side=position_side,
                        unrealized_pnl=unrealized_pnl,
                        entry_price=entry_price,
                        amount=amount,
                        leverage=Decimal(leverage)
                    )
                    self._perpetual_trading.set_leverage(trading_pair, leverage)
                    self._perpetual_trading.set_position(pos_key, _position)
                else:
                    self._perpetual_trading.remove_position(pos_key)

    async def _get_position_mode(self) -> Optional[PositionMode]:
        # NOTE: This is default to ONEWAY as there is nothing available on current version of Vega
        return self._position_mode

    async def _trading_pair_position_mode_set(self, mode: PositionMode, trading_pair: str) -> Tuple[bool, str]:
        # NOTE: There is no setting to set leverage in derive
        msg = "ok"
        success = True
        return success, msg

    async def _set_trading_pair_leverage(self, mode: PositionMode, trading_pair: str) -> Tuple[bool, str]:
        # NOTE: There is no setting to set leverage in derive
        msg = "ok"
        success = True
        return success, msg

    async def _fetch_last_fee_payment(self, trading_pair: str) -> Tuple[int, Decimal, Decimal]:
        symbol = await self.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
        payment_response = await self._api_post(
            path_url=CONSTANTS.GET_LAST_FUNDING_RATE_PATH_URL,
            data={
                "period": 3600,
                "page": 1,
                "page_size": 100,
                "start_timestamp": self._last_funding_time(),
                "instrument_name": symbol,
                "subaccount_id": self._sub_id
            },
            is_auth_required=True,
            limit_id=CONSTANTS.GET_LAST_FUNDING_RATE_PATH_URL)
        payload = {
            "instrument_name": symbol,
        }
        funding_info_response = await self._api_post(
            path_url=CONSTANTS.TICKER_PRICE_CHANGE_PATH_URL,
            data=payload)
        sorted_payment_response = payment_response["result"]["events"]
        if len(sorted_payment_response) < 1:
            timestamp, funding_rate, payment = 0, Decimal("-1"), Decimal("-1")
            return timestamp, funding_rate, payment
        funding_payment = sorted_payment_response[0]
        _payment = Decimal(funding_payment["funding"])
        funding_rate = Decimal(funding_info_response["result"]["perp_details"]["funding_rate"])
        timestamp = funding_payment["timestamp"] * 1e-3
        if _payment != Decimal("0"):
            payment = _payment
        else:
            timestamp, funding_rate, payment = 0, Decimal("-1"), Decimal("-1")
        return timestamp, funding_rate, payment

    def _last_funding_time(self) -> int:
        return int(((time.time() // 3600) - 1) * 3600 * 1e3)



import logging
from typing import List, Optional

from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.data_feed.candles_feed.candles_base import CandlesBase
from hummingbot.data_feed.candles_feed.gate_io_perpetual_candles import constants as CONSTANTS
from hummingbot.logger import HummingbotLogger


class GateioPerpetualCandles(CandlesBase):
    _logger: Optional[HummingbotLogger] = None

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self, trading_pair: str, interval: str = "1m", max_records: int = 150):
        super().__init__(trading_pair, interval, max_records)
        self.quanto_multiplier = None

    @property
    def name(self):
        return f"gate_io_perpetual_{self._trading_pair}"

    @property
    def rest_url(self):
        return CONSTANTS.REST_URL

    @property
    def wss_url(self):
        return CONSTANTS.WSS_URL

    @property
    def health_check_url(self):
        return self.rest_url + CONSTANTS.HEALTH_CHECK_ENDPOINT

    @property
    def candles_url(self):
        return self.rest_url + CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_endpoint(self):
        return CONSTANTS.CANDLES_ENDPOINT

    @property
    def candles_max_result_per_rest_request(self):
        return CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST

    @property
    def rate_limits(self):
        return CONSTANTS.RATE_LIMITS

    @property
    def intervals(self):
        return CONSTANTS.INTERVALS

    async def initialize_exchange_data(self):
        await self.get_exchange_trading_pair_quanto_multiplier()

    async def check_network(self) -> NetworkStatus:
        rest_assistant = await self._api_factory.get_rest_assistant()
        await rest_assistant.execute_request(url=self.health_check_url,
                                             throttler_limit_id=CONSTANTS.HEALTH_CHECK_ENDPOINT)
        return NetworkStatus.CONNECTED

    def get_exchange_trading_pair(self, trading_pair):
        return trading_pair.replace("-", "_")

    async def get_exchange_trading_pair_quanto_multiplier(self):
        rest_assistant = await self._api_factory.get_rest_assistant()
        data = await rest_assistant.execute_request(
            url=self.rest_url + CONSTANTS.CONTRACT_INFO_URL.format(contract=self._ex_trading_pair),
            throttler_limit_id=CONSTANTS.CONTRACT_INFO_URL
        )
        quanto_multiplier = float(data.get("quanto_multiplier"))
        self.quanto_multiplier = quanto_multiplier
        return quanto_multiplier

    @property
    def _is_first_candle_not_included_in_rest_request(self):
        return False

    @property
    def _is_last_candle_not_included_in_rest_request(self):
        return False

    def _get_rest_candles_params(self,
                                 start_time: Optional[int] = None,
                                 end_time: Optional[int] = None,
                                 limit: Optional[int] = CONSTANTS.MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST) -> dict:
        """
        For API documentation, please refer to:
        https://www.gate.io/docs/developers/apiv4/#get-futures-candlesticks
        """
        return {
            "contract": self._ex_trading_pair,
            "interval": self.interval,
            "from": start_time,
            "to": end_time
        }

    def _parse_rest_candles(self, data: dict, end_time: Optional[int] = None) -> List[List[float]]:
        new_hb_candles = []
        for i in data:
            timestamp = i.get("t")
            open = i.get("o")
            high = i.get("h")
            low = i.get("l")
            close = i.get("c")
            volume = i.get("v") * self.quanto_multiplier
            quote_asset_volume = i.get("sum")
            n_trades = 0
            taker_buy_base_volume = 0
            taker_buy_quote_volume = 0
            new_hb_candles.append([self.ensure_timestamp_in_seconds(timestamp), open, high, low, close, volume,
                                   quote_asset_volume, n_trades, taker_buy_base_volume, taker_buy_quote_volume])
        return new_hb_candles

    def ws_subscription_payload(self):
        return {
            "time": int(self._time()),
            "channel": CONSTANTS.WS_CANDLES_ENDPOINT,
            "event": "subscribe",
            "payload": [self.interval, self._ex_trading_pair]
        }

    def _parse_websocket_message(self, data: dict):
        candles_row_dict = {}
        if data.get("event") == "update" and data.get("channel") == "futures.candlesticks":
            for i in data["result"]:
                candles_row_dict["timestamp"] = self.ensure_timestamp_in_seconds(i["t"])
                candles_row_dict["open"] = i["o"]
                candles_row_dict["high"] = i["h"]
                candles_row_dict["low"] = i["l"]
                candles_row_dict["close"] = i["c"]
                candles_row_dict["volume"] = i["v"] * self.quanto_multiplier
                candles_row_dict["quote_asset_volume"] = i.get("sum", 0)
                candles_row_dict["n_trades"] = 0
                candles_row_dict["taker_buy_base_volume"] = 0
                candles_row_dict["taker_buy_quote_volume"] = 0
            return candles_row_dict



from typing import Any, Dict

# A single source of truth for constant variables related to the exchange
from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

# The max size of a digest is 66 characters (Vertex uses digests comprable to client order id).
MAX_ORDER_ID_LEN = 66

HEARTBEAT_TIME_INTERVAL = 30.0

ORDER_BOOK_DEPTH = 100

VERSION = "0.0.1"

EXCHANGE_NAME = "vertex"

DEFAULT_DOMAIN = "vertex"
TESTNET_DOMAIN = "vertex_testnet"

QUOTE = "USDC"

BASE_URLS = {
    DEFAULT_DOMAIN: "https://gateway.prod.vertexprotocol.com/v1",
    TESTNET_DOMAIN: "https://gateway.sepolia-test.vertexprotocol.com/v1",
}

WSS_URLS = {
    DEFAULT_DOMAIN: "wss://gateway.prod.vertexprotocol.com/v1/ws",
    TESTNET_DOMAIN: "wss://gateway.sepolia-test.vertexprotocol.com/v1/ws",
}

ARCHIVE_INDEXER_URLS = {
    DEFAULT_DOMAIN: "https://archive.prod.vertexprotocol.com/v1",
    TESTNET_DOMAIN: "https://archive.sepolia-test.vertexprotocol.com/v1",
}

WS_SUBSCRIBE_URLS = {
    DEFAULT_DOMAIN: "wss://gateway.prod.vertexprotocol.com/v1/subscribe",
    TESTNET_DOMAIN: "wss://gateway.vertexprotocol-vertexprotocol.com/v1/subscribe",
}

CONTRACTS = {
    DEFAULT_DOMAIN: "0xbbee07b3e8121227afcfe1e2b82772246226128e",
    TESTNET_DOMAIN: "0x5956d6f55011678b2cab217cd21626f7668ba6c5",
}

CHAIN_IDS = {
    DEFAULT_DOMAIN: 42161,
    TESTNET_DOMAIN: 421613,
}

HBOT_BROKER_ID = ""

SIDE_BUY = "BUY"
SIDE_SELL = "SELL"

TIME_IN_FORCE_GTC = "GTC"  # Good till cancelled
TIME_IN_FORCE_IOC = "IOC"  # Immediate or cancel
TIME_IN_FORCE_FOK = "FOK"  # Fill or kill
TIME_IN_FORCE_POSTONLY = "POSTONLY"  # PostOnly

# API PATHS
POST_PATH_URL = "/execute"
QUERY_PATH_URL = "/query"
INDEXER_PATH_URL = "/indexer"
SYMBOLS_PATH_URL = "/symbols"

# POST METHODS
PLACE_ORDER_METHOD = "place_order"
PLACE_ORDER_METHOD_NO_LEVERAGE = "place_order_no_leverage"
CANCEL_ORDERS_METHOD = "cancel_orders"
CANCEL_ALL_METHOD = "cancel_product_orders"

# REST QUERY API TYPES
STATUS_REQUEST_TYPE = "status"
ORDER_REQUEST_TYPE = "order"
SUBACCOUNT_INFO_REQUEST_TYPE = "subaccount_info"
MARKET_LIQUIDITY_REQUEST_TYPE = "market_liquidity"
ALL_PRODUCTS_REQUEST_TYPE = "all_products"
MARKET_PRICE_REQUEST_TYPE = "market_price"
FEE_RATES_REQUEST_TYPE = "fee_rates"
CONTRACTS_REQUEST_TYPE = "contracts"
SUBACCOUNT_ORDERS_REQUEST_TYPE = "subaccount_orders"
MAX_WITHDRAWABLE_REQUEST_TYPE = "max_withdrawable"

# WS API ENDPOINTS
WS_SUBSCRIBE_METHOD = "subscribe"
TOB_TOPIC_EVENT_TYPE = "best_bid_offer"
POSITION_CHANGE_EVENT_TYPE = "position_change"
SNAPSHOT_EVENT_TYPE = "market_liquidity"
TRADE_EVENT_TYPE = "trade"
DIFF_EVENT_TYPE = "book_depth"
FILL_EVENT_TYPE = "fill"
POSITION_CHANGE_EVENT_TYPE = "position_change"

# Products
# NOTE: Index 7+ is only on testnet
PRODUCTS = {
    0: {
        "symbol": "USDC",
        "market": None,
        DEFAULT_DOMAIN: "0x0000000000000000000000000000000000000000",
        TESTNET_DOMAIN: "0x0000000000000000000000000000000000000000",
    },
    1: {
        "symbol": "wBTC",
        "market": "wBTC/USDC",
        DEFAULT_DOMAIN: "0x70e5911371472e406f1291c621d1c8f207764d73",
        TESTNET_DOMAIN: "0x939b0915f9c3b657b9e9a095269a0078dd587491",
    },
    2: {
        "symbol": "BTC-PERP",
        "market": "wBTC/USDC",
        DEFAULT_DOMAIN: "0xf03f457a30e598d5020164a339727ef40f2b8fbc",
        TESTNET_DOMAIN: "0x291b578ff99bfef1706a2018d9dfdd98773e4f3e",
    },
    3: {
        "symbol": "wETH",
        "market": "wETH/USDC",
        DEFAULT_DOMAIN: "0x1c6281a78aa0ed88949c319cba5f0f0de2ce8353",
        TESTNET_DOMAIN: "0x4008c7b762d7000034207bdef628a798065c3dcc",
    },
    4: {
        "symbol": "ETH-PERP",
        "market": "wETH/USDC",
        DEFAULT_DOMAIN: "0xfe653438a1a4a7f56e727509c341d60a7b54fa91",
        TESTNET_DOMAIN: "0xe5106c497f8398ee8d1d6d246f08c125245d19ff",
    },
    5: {
        "symbol": "ARB",
        "market": "ARB/USDC",
        DEFAULT_DOMAIN: "0xb6304e9a6ca241376a5fc9294daa8fca65ddcdcd",
        TESTNET_DOMAIN: "0x49eff6d3de555be7a039d0b86471e3cb454b35de",
    },
    6: {
        "symbol": "ARB-PERP",
        "market": "ARB/USDC",
        DEFAULT_DOMAIN: "0x01ec802ae0ab1b2cc4f028b9fe6eb954aef06ed1",
        TESTNET_DOMAIN: "0xc5f223f12d091fba16141d4eeb5d39c5e0e2577c",
    },
    # TESTNET
    7: {
        "symbol": "ARB2",
        "market": "ARB2/USDC",
        DEFAULT_DOMAIN: None,
        TESTNET_DOMAIN: "0xf9144ddc09bd6961cbed631f8be708d2d1e87f57",
    },
    8: {
        "symbol": "ARB-PERP2",
        "market": "ARB2/USDC",
        DEFAULT_DOMAIN: None,
        TESTNET_DOMAIN: "0xa0c85ffadceba288fbcba1dcb780956c01b25cdf",
    },
    9: {
        "symbol": "ARB-PERP2",
        "market": "ARB2/USDC",
        DEFAULT_DOMAIN: None,
        TESTNET_DOMAIN: "0xa0c85ffadceba288fbcba1dcb780956c01b25cdf",
    },
    10: {
        "symbol": "ARB-PERP2",
        "market": "ARB2/USDC",
        DEFAULT_DOMAIN: None,
        TESTNET_DOMAIN: "0xa0c85ffadceba288fbcba1dcb780956c01b25cdf",
    },
}

# OrderStates
ORDER_STATE = {
    "PendingNew": OrderState.PENDING_CREATE,
    "New": OrderState.OPEN,
    "Filled": OrderState.FILLED,
    "PartiallyFilled": OrderState.PARTIALLY_FILLED,
    "Canceled": OrderState.CANCELED,
    "Rejected": OrderState.FAILED,
}

# Any call increases call rate in ALL pool, so e.g. a query/execute call will contribute to both ALL and query/execute pools.
ALL_ENDPOINTS_LIMIT = "All"
RATE_LIMITS = [
    RateLimit(limit_id=ALL_ENDPOINTS_LIMIT, limit=600, time_interval=10),
    RateLimit(
        limit_id=INDEXER_PATH_URL, limit=60, time_interval=1, linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)]
    ),
    RateLimit(
        limit_id=STATUS_REQUEST_TYPE,
        limit=60,
        time_interval=1,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=ORDER_REQUEST_TYPE,
        limit=60,
        time_interval=1,
        # NOTE: No weight for weight of 1...
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=SUBACCOUNT_INFO_REQUEST_TYPE,
        limit=60,
        time_interval=10,
        weight=10,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=MARKET_LIQUIDITY_REQUEST_TYPE,
        limit=60,
        time_interval=1,
        # NOTE: No weight for weight of 1...
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=ALL_PRODUCTS_REQUEST_TYPE,
        limit=12,
        time_interval=1,
        weight=5,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=MARKET_PRICE_REQUEST_TYPE,
        limit=60,
        time_interval=1,
        # NOTE: No weight for weight of 1...
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=FEE_RATES_REQUEST_TYPE,
        limit=30,
        time_interval=1,
        weight=2,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=CONTRACTS_REQUEST_TYPE,
        limit=60,
        time_interval=1,
        weight=1,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=SUBACCOUNT_ORDERS_REQUEST_TYPE,
        limit=30,
        time_interval=1,
        weight=2,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=MAX_WITHDRAWABLE_REQUEST_TYPE,
        limit=120,
        time_interval=10,
        weight=5,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    # NOTE: For spot with no leverage, there are different limits.
    # Review https://vertex-protocol.gitbook.io/docs/developer-resources/api/websocket-rest-api/executes/place-order
    RateLimit(
        limit_id=PLACE_ORDER_METHOD,
        limit=10,
        time_interval=1,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    RateLimit(
        limit_id=PLACE_ORDER_METHOD_NO_LEVERAGE,
        limit=5,
        time_interval=10,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    # NOTE: We're only providing one (1) digest at a time currently.
    # https://vertex-protocol.gitbook.io/docs/developer-resources/api/websocket-rest-api/executes/cancel-orders
    RateLimit(
        limit_id=CANCEL_ORDERS_METHOD,
        limit=600,
        time_interval=1,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
    # NOTE: This isn't currently in use.
    # https://vertex-protocol.gitbook.io/docs/developer-resources/api/websocket-rest-api/executes/cancel-product-orders
    RateLimit(
        limit_id=CANCEL_ALL_METHOD,
        limit=2,
        time_interval=1,
        linked_limits=[LinkedLimitWeightPair(ALL_ENDPOINTS_LIMIT)],
    ),
]

"""
https://vertex-protocol.gitbook.io/docs/developer-resources/api/api-errors
"""
ERRORS: Dict[int, Any] = {
    1000: {
        "code": 1000,
        "error_value": "RateLimit",
        "description": "Too Many Requests: You have exceeded the rate limit. Please reduce your request frequency and try again later.",
        "message": "",
    },
    1001: {
        "code": 1001,
        "error_value": "BlacklistedAddress",
        "description": "This address has been blacklisted from accessing the sequencer due to a violation of the Terms of Service. If you believe this is an error, please contact the Vertex team for assistance.",
        "message": "",
    },
    1002: {
        "code": 1002,
        "error_value": "BlockedLocation",
        "description": "Access from your current location ({location}) is blocked. Please check your location and try again.",
        "message": "",
    },
    1003: {
        "code": 1003,
        "error_value": "BlockedSubdivision",
        "description": "Access from your current location ({location} - {subdivision}) is blocked. Please check your location and try again.",
        "message": "",
    },
}



from bidict import bidict

from hummingbot.core.api_throttler.data_types import LinkedLimitWeightPair, RateLimit

REST_URL = "https://api.btcmarkets.net"
HEALTH_CHECK_ENDPOINT = "/v3/time"
CANDLES_ENDPOINT = "/v3/markets/{market_id}/candles"

WSS_URL = None

INTERVALS = bidict(
    {
        "1m": "1m",
        "3m": "3m",
        "5m": "5m",
        "15m": "15m",
        "30m": "30m",
        "1h": "1h",
        "2h": "2h",
        "3h": "3h",
        "4h": "4h",
        "6h": "6h",
        "1d": "1d",
        "1w": "1w",
        "1M": "1mo",  # BTC Markets uses "1mo" for 1 month
    }
)

POLL_INTERVAL = 5.0  # seconds

MAX_RESULTS_PER_CANDLESTICK_REST_REQUEST = 1000

# Rate Limits based on BTC Markets documentation
# Using the same rate limits as defined in btc_markets_constants.py
MARKETS_URL = "/v3/markets"
SERVER_TIME_PATH_URL = "/v3/time"

RATE_LIMITS = [
    RateLimit(limit_id=MARKETS_URL, limit=150, time_interval=10),
    RateLimit(limit_id=SERVER_TIME_PATH_URL, limit=50, time_interval=10),
    RateLimit(limit_id=CANDLES_ENDPOINT, limit=50, time_interval=10, linked_limits=[LinkedLimitWeightPair("raw", 1)]),
    RateLimit(
        limit_id=HEALTH_CHECK_ENDPOINT, limit=50, time_interval=10, linked_limits=[LinkedLimitWeightPair("raw", 1)]
    ),
]



import re
from typing import List, Match, Optional, Pattern

# W{TOKEN} only applies to a few special tokens. It should NOT match all W-prefixed token names like WAVE or WOW.
CAPITAL_W_SYMBOLS_PATTERN = re.compile(r"^W(BTC|ETH|AVAX|ALBT|XRP|POL)")

# w{TOKEN} generally means a wrapped token on the Ethereum network. e.g. wNXM, wDGLD.
SMALL_W_SYMBOLS_PATTERN = re.compile(r"^w(\w+)")

# {TOKEN}.e generally means a wrapped token on the Avalanche network.
DOT_E_SYMBOLS_PATTERN = re.compile(r"(\w+)\.e$", re.IGNORECASE)

USD_EQUIVALANT_TOKENS = ["USC"]


def unwrap_token_symbol(on_chain_token_symbol: str) -> str:
    patterns: List[Pattern] = [
        CAPITAL_W_SYMBOLS_PATTERN,
        SMALL_W_SYMBOLS_PATTERN,
        DOT_E_SYMBOLS_PATTERN
    ]
    for p in patterns:
        m: Optional[Match] = p.search(on_chain_token_symbol)
        if m is not None:
            return m.group(1)

    if on_chain_token_symbol in USD_EQUIVALANT_TOKENS:
        on_chain_token_symbol = "USDT"
    return on_chain_token_symbol
